/***
*
*	Copyright (c) 1996-2002, Valve LLC. All rights reserved.
*	
*	This product contains software technology licensed from Id 
*	Software, Inc. ("Id Technology").  Id Technology (c) 1996 Id Software, Inc. 
*	All Rights Reserved.
*
*   Use, distribution, and modification of this source code and/or resulting
*   object code is restricted to non-commercial enhancements to products from
*   Valve LLC.  All other use, distribution, or modification is prohibited
*   without written permission from Valve LLC.
*
****/

#include "extdll.h"
#include "util.h"
#include "cbase.h"
#include "client.h"
#include "player.h"
#include "ent/func_break.h"

// =================== FUNC_WALL ==============================================

class CFuncWall : public CBaseDelay
{
	DECLARE_CLASS( CFuncWall, CBaseDelay );
public:
	void Spawn( void );
	virtual void TurnOn( void );
	virtual void TurnOff( void );
	void Use( CBaseEntity *pActivator, CBaseEntity *pCaller, USE_TYPE useType, float value );
	virtual int ObjectCaps( void ) { return CBaseEntity :: ObjectCaps() & ~FCAP_ACROSS_TRANSITION; }
};

LINK_ENTITY_TO_CLASS( func_wall, CFuncWall );

void CFuncWall :: Spawn( void )
{
	pev->flags |= FL_WORLDBRUSH;
	pev->solid = SOLID_BSP;
	pev->movetype = MOVETYPE_PUSH;
	m_iState = STATE_OFF;

	SetLocalAngles( g_vecZero );
	SET_MODEL( edict(), GetModel() );

	// LRC (support generic switchable texlight)
	if( m_iStyle >= 32 )
		LIGHT_STYLE( m_iStyle, "a" );
	else if( m_iStyle <= -32 )
		LIGHT_STYLE( -m_iStyle, "z" );

	if( m_hParent != NULL || FClassnameIs( pev, "func_wall_toggle" ))
		m_pUserData = WorldPhysic->CreateKinematicBodyFromEntity( this );
	else m_pUserData = WorldPhysic->CreateStaticBodyFromEntity( this );
}

void CFuncWall :: TurnOff( void )
{
	pev->frame = 0;
	m_iState = STATE_OFF;
}

void CFuncWall :: TurnOn( void )
{
	pev->frame = 1;
	m_iState = STATE_ON;
}

void CFuncWall :: Use( CBaseEntity *pActivator, CBaseEntity *pCaller, USE_TYPE useType, float value )
{
	if( IsLockedByMaster( pActivator ))
		return;

	if( ShouldToggle( useType ))
	{
		if( GetState() == STATE_ON )
			TurnOff();
		else TurnOn();

		// LRC (support generic switchable texlight)
		if( m_iStyle >= 32 )
		{
			if( pev->frame )
				LIGHT_STYLE( m_iStyle, "z" );
			else
				LIGHT_STYLE( m_iStyle, "a" );
		}
		else if( m_iStyle <= -32 )
		{
			if( pev->frame )
				LIGHT_STYLE( -m_iStyle, "a" );
			else
				LIGHT_STYLE( -m_iStyle, "z" );
		}
	}
}

// =================== FUNC_WALL_TOGGLE ==============================================

#define SF_WALL_START_OFF		BIT( 0 )

class CFuncWallToggle : public CFuncWall
{
	DECLARE_CLASS( CFuncWallToggle, CFuncWall );
public:
	void Spawn( void );
	void TurnOff( void );
	void TurnOn( void );
};

LINK_ENTITY_TO_CLASS( func_wall_toggle, CFuncWallToggle );

void CFuncWallToggle :: Spawn( void )
{
	BaseClass::Spawn();

	if( FBitSet( pev->spawnflags, SF_WALL_START_OFF ))
		TurnOff();
	else TurnOn();
}

void CFuncWallToggle :: TurnOff( void )
{
	WorldPhysic->EnableCollision( this, FALSE );
	pev->solid = SOLID_NOT;
	pev->effects |= EF_NODRAW;
	RelinkEntity( FALSE );
	m_iState = STATE_OFF;
}

void CFuncWallToggle :: TurnOn( void )
{
	WorldPhysic->EnableCollision( this, TRUE );
	pev->solid = SOLID_BSP;
	pev->effects &= ~EF_NODRAW;
	RelinkEntity( FALSE );
	m_iState = STATE_ON;
}

// =================== FUNC_CONVEYOR ==============================================

#define SF_CONVEYOR_VISUAL		BIT( 0 )
#define SF_CONVEYOR_NOTSOLID		BIT( 1 )
#define SF_CONVEYOR_STARTOFF		BIT( 2 )

class CFuncConveyor : public CBaseDelay
{
	DECLARE_CLASS( CFuncConveyor, CBaseDelay );
public:
	void Spawn( void );
	void Use( CBaseEntity *pActivator, CBaseEntity *pCaller, USE_TYPE useType, float value );
	virtual int ObjectCaps( void ) { return (BaseClass::ObjectCaps() & ~FCAP_ACROSS_TRANSITION) | FCAP_SET_MOVEDIR; }
	void UpdateSpeed( float speed );

	DECLARE_DATADESC();

	float	m_flMaxSpeed;
};

LINK_ENTITY_TO_CLASS( func_conveyor, CFuncConveyor );

BEGIN_DATADESC( CFuncConveyor )
	DEFINE_FIELD( m_flMaxSpeed, FIELD_FLOAT ),
END_DATADESC()

void CFuncConveyor :: Spawn( void )
{
	pev->flags |= FL_WORLDBRUSH;
	pev->solid = SOLID_BSP;
	pev->movetype = MOVETYPE_PUSH;

	SET_MODEL( edict(), GetModel() );

	if( !FBitSet( pev->spawnflags, SF_CONVEYOR_VISUAL ))
		SetBits( pev->flags, FL_CONVEYOR );

	// is mapper forgot set angles?
	if( pev->movedir == g_vecZero )
		pev->movedir = Vector( 1, 0, 0 );

	// HACKHACK - This is to allow for some special effects
	if( FBitSet( pev->spawnflags, SF_CONVEYOR_NOTSOLID ))
	{
		pev->solid = SOLID_NOT;
		pev->skin = 0; // don't want the engine thinking we've got special contents on this brush
	}
	else
	{
		if( m_hParent )
			m_pUserData = WorldPhysic->CreateKinematicBodyFromEntity( this );
		else m_pUserData = WorldPhysic->CreateStaticBodyFromEntity( this );
	}

	if( pev->speed == 0 )
		pev->speed = 100;

	m_flMaxSpeed = pev->speed;	// save initial speed

	if( FBitSet( pev->spawnflags, SF_CONVEYOR_STARTOFF ))
		UpdateSpeed( 0 );
	else UpdateSpeed( m_flMaxSpeed );
}

void CFuncConveyor :: UpdateSpeed( float speed )
{
	// make sure the sign is correct - positive for forward rotation,
	// negative for reverse rotation.
	speed = fabs( speed );
	if( pev->impulse ) speed *= -1;

	// encode it as an integer with 4 fractional bits
	int speedCode = (int)(fabs( speed ) * 16.0f);

	// HACKHACK -- This is ugly, but encode the speed in the rendercolor
	// to avoid adding more data to the network stream
	pev->rendercolor.x = (speed < 0) ? 1 : 0;
	pev->rendercolor.y = (speedCode >> 8);
	pev->rendercolor.z = (speedCode & 0xFF);

	// set conveyor state
	m_iState = ( speed != 0 ) ? STATE_ON : STATE_OFF;
	pev->speed = speed;	// update physical speed too
}

void CFuncConveyor :: Use( CBaseEntity *pActivator, CBaseEntity *pCaller, USE_TYPE useType, float value )
{
	if( IsLockedByMaster( pActivator ))
		return;

	if( ShouldToggle( useType ))
	{
		if( useType == USE_SET )
		{
			if( value != 0 ) 
			{
				pev->impulse = value < 0 ? true : false;
				value = fabs( value );
				pev->dmg = bound( 0.1, value, 1 ) * m_flMaxSpeed;
			}
			else UpdateSpeed( 0 ); // stop
			return;
		}
		else if( useType == USE_RESET )
		{
			// restore last speed
			UpdateSpeed( pev->dmg );
			return;
		}
  
		pev->impulse = !pev->impulse;
		UpdateSpeed( m_flMaxSpeed );
	}
}

// =================== FUNC_TRANSPORTER ==============================================
// like conveyor but with more intuitive controls

#define SF_TRANSPORTER_STARTON	BIT( 0 )
#define SF_TRANSPORTER_SOLID		BIT( 1 )

class CFuncTransporter : public CBaseDelay
{
	DECLARE_CLASS( CFuncTransporter, CBaseDelay );
public:
	void Spawn( void );
	void Use( CBaseEntity *pActivator, CBaseEntity *pCaller, USE_TYPE useType, float value );
	virtual int ObjectCaps( void ) { return (BaseClass::ObjectCaps() & ~FCAP_ACROSS_TRANSITION) | FCAP_SET_MOVEDIR; }
	void UpdateSpeed( float speed );
	void Think( void );

	DECLARE_DATADESC();

	float	m_flMaxSpeed;
	float	m_flWidth;
};

LINK_ENTITY_TO_CLASS( func_transporter, CFuncTransporter );

BEGIN_DATADESC( CFuncTransporter )
	DEFINE_FIELD( m_flMaxSpeed, FIELD_FLOAT ),
	DEFINE_FIELD( m_flWidth, FIELD_FLOAT ),
END_DATADESC()

void CFuncTransporter :: Spawn( void )
{
	pev->flags |= FL_WORLDBRUSH;
	pev->movetype = MOVETYPE_PUSH;
	pev->solid = SOLID_BSP;

	SET_MODEL( edict(), GetModel() );
	SetBits( pev->effects, EF_CONVEYOR );

	// is mapper forgot set angles?
	if( pev->movedir == g_vecZero )
		pev->movedir = Vector( 1, 0, 0 );

	model_t *bmodel;

	// get a world struct
	if(( bmodel = (model_t *)MODEL_HANDLE( pev->modelindex )) == NULL )
	{
		ALERT( at_warning, "%s: unable to fetch model pointer %s\n", GetDebugName(), GetModel( ));
		m_flWidth = 64.0f;	// default
	}
	else
	{
		// lookup for the scroll texture
		for( int i = 0; i < bmodel->nummodelsurfaces; i++ )
		{
			msurface_t *face = &bmodel->surfaces[bmodel->firstmodelsurface + i];
			if( FBitSet( face->flags, SURF_CONVEYOR ))
			{
				m_flWidth = face->texinfo->texture->width;
				break;
			}
		}

		if( m_flWidth <= 0.0 )
		{
			ALERT( at_warning, "%s: unable to find scrolling texture\n", GetDebugName() );
			m_flWidth = 64.0f;	// default
		}
	}

	if( FBitSet( pev->spawnflags, SF_TRANSPORTER_SOLID ))
	{
		if( m_hParent )
			m_pUserData = WorldPhysic->CreateKinematicBodyFromEntity( this );
		else m_pUserData = WorldPhysic->CreateStaticBodyFromEntity( this );
		SetBits( pev->flags, FL_CONVEYOR );
	}
	else
	{
		pev->solid = SOLID_NOT;
		pev->skin = 0; // don't want the engine thinking we've got special contents on this brush
	}

	if( pev->speed == 0 )
		pev->speed = 100;

	m_flMaxSpeed = pev->speed;	// save initial speed

	if( FBitSet( pev->spawnflags, SF_TRANSPORTER_STARTON ))
		UpdateSpeed( m_flMaxSpeed );
	else UpdateSpeed( 0 );
	SetNextThink( 0 );
}

void CFuncTransporter :: UpdateSpeed( float speed )
{
	// set conveyor state
	m_iState = ( speed != 0 ) ? STATE_ON : STATE_OFF;
	pev->speed = speed;
}

void CFuncTransporter :: Use( CBaseEntity *pActivator, CBaseEntity *pCaller, USE_TYPE useType, float value )
{
	if( IsLockedByMaster( pActivator ))
		return;

	if( ShouldToggle( useType ))
	{
		if( useType == USE_SET )
		{
			UpdateSpeed( value );
		}
		else
		{
			if( GetState() == STATE_ON )
				UpdateSpeed( 0 );
			else UpdateSpeed( m_flMaxSpeed );
		}
	}
}

void CFuncTransporter :: Think( void )
{
	if( pev->speed != 0 )
	{
		pev->fuser1 += pev->speed * gpGlobals->frametime * (1.0f / m_flWidth); 

		// don't exceed some limits to prevent precision loosing
		while( pev->fuser1 > m_flWidth )
			pev->fuser1 -= m_flWidth;
		while( pev->fuser1 < -m_flWidth )
			pev->fuser1 += m_flWidth;
	}

	// think every frame
	SetNextThink( gpGlobals->frametime );
}

// =================== FUNC_ILLUSIONARY ==============================================

class CFuncIllusionary : public CBaseEntity
{
	DECLARE_CLASS( CFuncIllusionary, CBaseEntity );
public:
	virtual int ObjectCaps( void ) { return BaseClass :: ObjectCaps() & ~FCAP_ACROSS_TRANSITION; }
	void Spawn( void );
};

LINK_ENTITY_TO_CLASS( func_illusionary, CFuncIllusionary );

void CFuncIllusionary :: Spawn( void )
{
	pev->movetype = MOVETYPE_NONE;  
	pev->solid = SOLID_NOT; // always solid_not 
	SET_MODEL( edict(), GetModel() );
	
	// I'd rather eat the network bandwidth of this than figure out how to save/restore
	// these entities after they have been moved to the client, or respawn them ala Quake
	// Perhaps we can do this in deathmatch only.
	// MAKE_STATIC( edict() );
}

// =================== FUNC_MONSTERCLIP ==============================================

class CFuncMonsterClip : public CBaseEntity
{
	DECLARE_CLASS( CFuncMonsterClip, CBaseEntity );
public:
	virtual int ObjectCaps( void ) { return BaseClass :: ObjectCaps() & ~FCAP_ACROSS_TRANSITION; }
	void Spawn( void );
};

LINK_ENTITY_TO_CLASS( func_monsterclip, CFuncMonsterClip );

void CFuncMonsterClip :: Spawn( void )
{
	pev->solid = SOLID_BSP;
	pev->movetype = MOVETYPE_PUSH;
	SetBits( pev->flags, FL_MONSTERCLIP );
	SetBits( m_iFlags, MF_TRIGGER );
	pev->effects |= EF_NODRAW;

	// link into world
	SET_MODEL( edict(), GetModel() );
}



// =================== FUNC_PENDULUM ==============================================

#define SF_PENDULUM_START_ON		BIT( 0 )
#define SF_PENDULUM_SWING		BIT( 1 )
#define SF_PENDULUM_PASSABLE		BIT( 3 )
#define SF_PENDULUM_AUTO_RETURN	BIT( 4 )

class CPendulum : public CBaseDelay
{
	DECLARE_CLASS( CPendulum, CBaseDelay );
public:
	void	Spawn( void );
	void	KeyValue( KeyValueData *pkvd );
	void	Swing( void );
	void	Stop( void );
	void	RopeTouch( CBaseEntity *pOther ); // this touch func makes the pendulum a rope
	virtual int ObjectCaps( void ) { return (BaseClass :: ObjectCaps() & ~FCAP_ACROSS_TRANSITION) | FCAP_HOLD_ANGLES; }
	void	Use( CBaseEntity *pActivator, CBaseEntity *pCaller, USE_TYPE useType, float value );
	void	Touch( CBaseEntity *pOther );
	void	Blocked( CBaseEntity *pOther );

	DECLARE_DATADESC();
	
	float	m_accel;	// Acceleration
	float	m_distance;
	float	m_time;
	float	m_damp;
	float	m_maxSpeed;
	float	m_dampSpeed;
	Vector	m_center;
	Vector	m_start;
};

LINK_ENTITY_TO_CLASS( func_pendulum, CPendulum );

BEGIN_DATADESC( CPendulum )
	DEFINE_FIELD( m_accel, FIELD_FLOAT ),
	DEFINE_KEYFIELD( m_distance, FIELD_FLOAT, "distance" ),
	DEFINE_FIELD( m_time, FIELD_TIME ),
	DEFINE_FIELD( m_damp, FIELD_FLOAT ),
	DEFINE_FIELD( m_maxSpeed, FIELD_FLOAT ),
	DEFINE_FIELD( m_dampSpeed, FIELD_FLOAT ),
	DEFINE_FIELD( m_center, FIELD_VECTOR ),
	DEFINE_FIELD( m_start, FIELD_VECTOR ),
	DEFINE_FUNCTION( Swing ),
	DEFINE_FUNCTION( Stop ),
	DEFINE_FUNCTION( RopeTouch ),
END_DATADESC()

void CPendulum :: KeyValue( KeyValueData *pkvd )
{
	if( FStrEq( pkvd->szKeyName, "distance" ))
	{
		m_distance = Q_atof( pkvd->szValue );
		pkvd->fHandled = TRUE;
	}
	else if( FStrEq( pkvd->szKeyName, "damp" ))
	{
		m_damp = Q_atof( pkvd->szValue ) * 0.001;
		pkvd->fHandled = TRUE;
	}
	else CBaseDelay::KeyValue( pkvd );
}

void CPendulum :: Spawn( void )
{
	// set the axis of rotation
	CBaseToggle::AxisDir( pev );

	if( FBitSet( pev->spawnflags, SF_PENDULUM_PASSABLE ))
		pev->solid = SOLID_NOT;
	else pev->solid = SOLID_BSP;

	pev->movetype = MOVETYPE_PUSH;

	SET_MODEL( edict(), GetModel() );

	if( m_distance == 0 )
	{
		ALERT( at_error, "%s [%i] has invalid distance\n", GetClassname(), entindex());
		return;
	}

	if( pev->speed == 0 )
		pev->speed = 100;

	SetLocalAngles( m_vecTempAngles );

	// calculate constant acceleration from speed and distance
	m_accel = (pev->speed * pev->speed) / ( 2 * fabs( m_distance ));
	m_maxSpeed = pev->speed;
	m_start = GetLocalAngles();
	m_center = GetLocalAngles() + ( m_distance * 0.5f ) * pev->movedir;

	m_pUserData = WorldPhysic->CreateKinematicBodyFromEntity( this );

	if( FBitSet( pev->spawnflags, SF_PENDULUM_START_ON ))
	{		
		SetThink( SUB_CallUseToggle );
		SetNextThink( 0.1 );
	}
	pev->speed = 0;

	if( FBitSet( pev->spawnflags, SF_PENDULUM_SWING ))
	{
		SetTouch ( RopeTouch );
	}
}

void CPendulum :: Use( CBaseEntity *pActivator, CBaseEntity *pCaller, USE_TYPE useType, float value )
{
	if( IsLockedByMaster( pActivator ) || m_distance == 0 )
		return;

	if( pev->speed )
	{
		// pendulum is moving, stop it and auto-return if necessary
		if( FBitSet( pev->spawnflags, SF_PENDULUM_AUTO_RETURN ))
		{		
			float delta = CBaseToggle::AxisDelta( pev->spawnflags, GetLocalAngles(), m_start );

			SetLocalAvelocity( m_maxSpeed * pev->movedir );
			SetNextThink( delta / m_maxSpeed );
			SetThink( Stop );
		}
		else
		{
			pev->speed = 0; // dead stop
			SetThink( NULL );
			SetLocalAvelocity( g_vecZero );
			m_iState = STATE_OFF;
		}
	}
	else
	{
		SetNextThink( 0.1f );		// start the pendulum moving
		SetThink( Swing );
		m_time = gpGlobals->time;		// save time to calculate dt
		m_dampSpeed = m_maxSpeed;
	}
}

void CPendulum :: Stop( void )
{
	SetLocalAngles( m_start );
	pev->speed = 0; // dead stop
	SetThink( NULL );
	SetLocalAvelocity( g_vecZero );
	m_iState = STATE_OFF;
}

void CPendulum::Blocked( CBaseEntity *pOther )
{
	m_time = gpGlobals->time;
}

void CPendulum :: Swing( void )
{
	float delta, dt;
	
	delta = CBaseToggle::AxisDelta( pev->spawnflags, GetLocalAngles(), m_center );

	dt = gpGlobals->time - m_time;	// how much time has passed?
	m_time = gpGlobals->time;		// remember the last time called

	// integrate velocity
	if( delta > 0 && m_accel > 0 )
		pev->speed -= m_accel * dt;
	else pev->speed += m_accel * dt;

	pev->speed = bound( -m_maxSpeed, pev->speed, m_maxSpeed );

	m_iState = STATE_ON;

	// scale the destdelta vector by the time spent traveling to get velocity
	SetLocalAvelocity( pev->speed * pev->movedir );

	// call this again
	SetNextThink( 0.1f );
	SetMoveDoneTime( 0.1f );

	if( m_damp )
	{
		m_dampSpeed -= m_damp * m_dampSpeed * dt;

		if( m_dampSpeed < 30.0 )
		{
			SetLocalAngles( m_center );
			pev->speed = 0;
			SetThink( NULL );
			SetLocalAvelocity( g_vecZero );
			m_iState = STATE_OFF;
		}
		else if( pev->speed > m_dampSpeed )
		{
			pev->speed = m_dampSpeed;
		}
		else if( pev->speed < -m_dampSpeed )
		{
			pev->speed = -m_dampSpeed;
		}
	}
}

void CPendulum :: Touch ( CBaseEntity *pOther )
{
	if( pev->dmg <= 0 )
		return;

	// we can't hurt this thing, so we're not concerned with it
	if( !pOther->pev->takedamage )
		return;

	// calculate damage based on rotation speed
	float damage = pev->dmg * pev->speed * 0.01;

	if( damage < 0 ) damage = -damage;

	pOther->TakeDamage( pev, pev, damage, DMG_CRUSH );

	Vector vNewVel = (pOther->GetAbsOrigin( ) - GetAbsOrigin( )).Normalize();
	pOther->SetAbsVelocity( vNewVel * damage );
}

void CPendulum :: RopeTouch ( CBaseEntity *pOther )
{
	if( !pOther->IsPlayer( ))
	{
		// not a player!
		ALERT( at_console, "Not a client\n" );
		return;
	}

	if( pOther->edict() == pev->enemy )
	{
		// this player already on the rope.
		return;
	}

	pev->enemy = pOther->edict();
	pOther->SetAbsVelocity( g_vecZero );
	pOther->pev->movetype = MOVETYPE_NONE;
}

// =================== FUNC_CLOCK ==============================================

#define SECONDS_PER_MINUTE	60
#define SECONDS_PER_HOUR	3600
#define SECODNS_PER_DAY	43200

class CFuncClock : public CBaseDelay
{
	DECLARE_CLASS( CFuncClock, CBaseDelay );
public:
	void Spawn ( void );
	void Think( void );
	void Activate( void );
	void KeyValue( KeyValueData *pkvd );
	virtual int ObjectCaps( void ) { return BaseClass :: ObjectCaps() & ~FCAP_ACROSS_TRANSITION; }

	DECLARE_DATADESC();
private:
	Vector	m_vecCurtime;
	Vector	m_vecFinaltime;
	float	m_flCurTime;	// current unconverted time in seconds
	int	m_iClockType;	// also contain seconds count for different modes
	int	m_iHoursCount;
	BOOL	m_fInit;		// this clock already init
};

LINK_ENTITY_TO_CLASS( func_clock, CFuncClock );

BEGIN_DATADESC( CFuncClock )
	DEFINE_FIELD( m_vecCurtime, FIELD_VECTOR ),
	DEFINE_FIELD( m_vecFinaltime, FIELD_VECTOR ),
	DEFINE_FIELD( m_flCurTime, FIELD_VECTOR ),
	DEFINE_FIELD( m_iClockType, FIELD_INTEGER ),
	DEFINE_FIELD( m_iHoursCount, FIELD_INTEGER ),
	DEFINE_FIELD( m_fInit, FIELD_BOOLEAN ),
END_DATADESC()

void CFuncClock :: KeyValue( KeyValueData *pkvd )
{
	if( FStrEq( pkvd->szKeyName, "type" ))
	{
		switch( Q_atoi( pkvd->szValue ))
		{
		case 1:
			m_iClockType = SECONDS_PER_HOUR;
			break;
		case 2:
			m_iClockType = SECODNS_PER_DAY;
			break;
		default:
			m_iClockType = SECONDS_PER_MINUTE;
			break;
		}
		pkvd->fHandled = TRUE;
	}
	else if( FStrEq( pkvd->szKeyName, "curtime" ))
	{
		m_vecCurtime = Q_atov( pkvd->szValue );
		pkvd->fHandled = TRUE;
	}
	else if( FStrEq( pkvd->szKeyName, "event" ))
	{
		pev->netname = ALLOC_STRING( pkvd->szValue );
		pkvd->fHandled = TRUE;
	}
	else CBaseEntity::KeyValue( pkvd );
}

void CFuncClock :: Spawn ( void )
{
	CBaseToggle::AxisDir( pev );

	m_iState = STATE_ON;
	pev->solid = SOLID_NOT;

	SET_MODEL( edict(), GetModel() );
	
	if( m_iClockType == SECODNS_PER_DAY )
	{
		// normalize our time
		if( m_vecCurtime.x > 11 ) m_vecCurtime.x = 0;
		if( m_vecCurtime.y > 59 ) m_vecCurtime.y = 0;
		if( m_vecCurtime.z > 59 ) m_vecCurtime.z = 0;
		
		// member full hours
		m_iHoursCount = m_vecCurtime.x;
		
		// calculate seconds
		m_vecFinaltime.z = m_vecCurtime.z * ( SECONDS_PER_MINUTE / 60 );
                    m_vecFinaltime.y = m_vecCurtime.y * ( SECONDS_PER_HOUR / 60 ) + m_vecFinaltime.z;
                    m_vecFinaltime.x = m_vecCurtime.x * ( SECODNS_PER_DAY / 12 ) + m_vecFinaltime.y;
	}
}

void CFuncClock :: Activate( void )
{
	if( m_fInit ) return;

	if( m_iClockType == SECODNS_PER_DAY && m_vecCurtime != g_vecZero )
	{
		// try to find minutes and seconds entity
		CBaseEntity *pEntity = NULL;
		while( pEntity = UTIL_FindEntityInSphere( pEntity, GetLocalOrigin(), pev->size.z ))
		{
			if( FClassnameIs( pEntity, "func_clock" ))
			{
				CFuncClock *pClock = (CFuncClock *)pEntity;
				// write start hours, minutes and seconds
				switch( pClock->m_iClockType )
				{
				case SECODNS_PER_DAY:
					// NOTE: here we set time for himself through FindEntityInSphere
					pClock->m_flCurTime = m_vecFinaltime.x;
					break;
				case SECONDS_PER_HOUR:
					pClock->m_flCurTime = m_vecFinaltime.y;
					break;
				default:
					pClock->m_flCurTime = m_vecFinaltime.z;
					break;
				}
			}
		}
	}

	// clock start	
	SetNextThink( 0 );
	m_fInit = 1;
}

void CFuncClock :: Think( void )
{
	float seconds, ang, pos;

	seconds = gpGlobals->time + m_flCurTime;
	pos = seconds / m_iClockType;
	pos = pos - floor( pos );
	ang = 360 * pos;

	SetLocalAngles( pev->movedir * ang );

	if( m_iClockType == SECODNS_PER_DAY )
	{
		int hours = GetLocalAngles().Length() / 30;
		if( m_iHoursCount != hours )
		{
			// member new hour
			m_iHoursCount = hours;
			if( hours == 0 ) hours = 12;	// merge for 0.00.00

			// send hours info
			UTIL_FireTargets( pev->netname, this, this, USE_SET, hours );
			UTIL_FireTargets( pev->netname, this, this, USE_ON );
		}
	}

	RelinkEntity( FALSE );

	// set clock resolution
	SetNextThink( 1.0f );
}

// =================== FUNC_LIGHT ==============================================

#define SF_LIGHT_START_ON		BIT( 0 )

class CFuncLight : public CBreakable
{
	DECLARE_CLASS( CFuncLight, CBreakable );
public:
	void Spawn( void );
	void Use( CBaseEntity *pActivator, CBaseEntity *pCaller, USE_TYPE useType, float value );
	virtual int TakeDamage( entvars_t* pevInflictor, entvars_t* pevAttacker, float flDamage, int bitsDamageType );
	void TraceAttack( entvars_t *pevAttacker, float flDamage, Vector vecDir, TraceResult *ptr, int bitsDamageType );
	void Flicker( void );
	void Die( void );

	DECLARE_DATADESC();
private:
	int	m_iFlickerMode;
	Vector	m_vecLastDmgPoint;
	float	m_flNextFlickerTime;
};
LINK_ENTITY_TO_CLASS( func_light, CFuncLight );

BEGIN_DATADESC( CFuncLight )
	DEFINE_FIELD( m_iFlickerMode, FIELD_INTEGER ),
	DEFINE_FIELD( m_flNextFlickerTime, FIELD_FLOAT ),
	DEFINE_FIELD( m_vecLastDmgPoint, FIELD_POSITION_VECTOR ),
	DEFINE_FUNCTION( Flicker ),
END_DATADESC()

void CFuncLight :: Spawn( void )
{	
	m_Material = matGlass;

	CBreakable::Spawn();
	SET_MODEL( edict(), GetModel() );

	// probably map compiler haven't func_light support
	if( m_iStyle <= 0 || m_iStyle >= 256 )
	{
		if( GetTargetname()[0] )
			ALERT( at_error, "%s with name %s has bad lightstyle %i. Disabled\n", GetClassname(), GetTargetname(), m_iStyle );
		else ALERT( at_error, "%s [%i] has bad lightstyle %i. Disabled\n", GetClassname(), entindex(), m_iStyle );
		m_iState = STATE_DEAD; // lamp is dead
	}
	
	if( FBitSet( pev->spawnflags, SF_LIGHT_START_ON ))
		Use( this, this, USE_ON, 0 );
	else Use( this, this, USE_OFF, 0 );

	if( pev->health <= 0 )
		pev->takedamage = DAMAGE_NO;
	else pev->takedamage = DAMAGE_YES;
}

void CFuncLight :: Use( CBaseEntity *pActivator, CBaseEntity *pCaller, USE_TYPE useType, float value )
{
	m_hActivator = pActivator;
	
	if( IsLockedByMaster( ))
		return;

	if( m_iState == STATE_DEAD )
		return; // lamp is broken

	if( useType == USE_TOGGLE )
	{
		if( m_iState == STATE_OFF )
			useType = USE_ON;
		else useType = USE_OFF;
	}

	if( useType == USE_ON )
	{
		if( m_flDelay )
		{
			// make flickering delay
			m_iState = STATE_TURN_ON;
			LIGHT_STYLE( m_iStyle, "mmamammmmammamamaaamammma" );
			pev->frame = 0; // light texture is on
			SetThink( Flicker );
			SetNextThink( m_flDelay );
		}
		else
		{         // instant enable
			m_iState = STATE_ON;
			LIGHT_STYLE( m_iStyle, "k" );
			pev->frame = 0; // light texture is on
			UTIL_FireTargets( pev->target, this, this, USE_ON );
		}
	}
	else if( useType == USE_OFF )
	{
		LIGHT_STYLE( m_iStyle, "a" );
		UTIL_FireTargets( pev->target, this, this, USE_OFF );
		pev->frame = 1;// light texture is off
		m_iState = STATE_OFF;
	}
	else if( useType == USE_SET )
	{
		// a script die (dramatic effect)
		Die();
	}
}

void CFuncLight :: TraceAttack( entvars_t *pevAttacker, float flDamage, Vector vecDir, TraceResult *ptr, int bitsDamageType )
{
	if( pev->takedamage == DAMAGE_NO )
		return;

	if( m_iState == STATE_DEAD )
		return;

	const char *pTextureName = NULL;
	Vector start = pevAttacker->origin + pevAttacker->view_ofs;
	Vector end = start + vecDir * 1024;
	edict_t *pHit = ptr->pHit;

	if( pHit )
		pTextureName = TRACE_TEXTURE( pHit, start, end );

	if( pTextureName != NULL && ( !Q_strncmp( pTextureName, "+0~", 3 ) || *pTextureName == '~' ))
	{
		// take damage only at light texture
		UTIL_Sparks( ptr->vecEndPos );
		m_vecLastDmgPoint = ptr->vecEndPos;
	}

	CBreakable::TraceAttack( pevAttacker, flDamage, vecDir, ptr, bitsDamageType );
}

int CFuncLight :: TakeDamage( entvars_t* pevInflictor, entvars_t* pevAttacker, float flDamage, int bitsDamageType )
{
	if( bitsDamageType & DMG_BLAST )
		flDamage *= 3.0f;
	else if( bitsDamageType & DMG_CLUB )
		flDamage *= 2.0f;
	else if( bitsDamageType & DMG_SONIC )
		flDamage *= 1.2f;
	else if( bitsDamageType & DMG_SHOCK )
		flDamage *= 10.0f;//!!!! over voltage
	else if( bitsDamageType & DMG_BULLET )
		flDamage *= 0.5f; // half damage at bullet

	pev->health -= flDamage;

	if( pev->health <= 0.0f )
	{
		Die();
		return 0;
	}

	CBreakable::DamageSound();

	return 1;
}

void CFuncLight :: Die( void )
{
	// lamp is random choose die style
	if( m_iState == STATE_OFF )
	{
		pev->frame = 1; // light texture is off
		LIGHT_STYLE( m_iStyle, "a" );
		DontThink();
	}
	else
	{         // simple randomization
		m_iFlickerMode = RANDOM_LONG( 1, 2 );
		SetThink( Flicker );
		SetNextThink( 0.1f + RANDOM_LONG( 0.1f, 0.2f ));
	}

	m_iState = STATE_DEAD;
	pev->health = 0;
	pev->takedamage = DAMAGE_NO;
	UTIL_FireTargets( pev->target, this, this, USE_OFF );

	switch( RANDOM_LONG( 0, 1 ))
	{
	case 0:
		EMIT_SOUND( edict(), CHAN_VOICE, "debris/bustglass1.wav", 0.7, ATTN_IDLE );
		break;
	case 1:
		EMIT_SOUND( edict(), CHAN_VOICE, "debris/bustglass2.wav", 0.8, ATTN_IDLE );
		break;
	}

	Vector vecSpot = GetAbsOrigin() + (pev->mins + pev->maxs) * 0.5f;

Vector vecSize;

vecSize.x=pev->size.x;
vecSize.y=pev->size.y;
vecSize.z=pev->size.z;

UTIL_Debris(vecSize, 10, vecSpot, m_idShard, Vector(0,0,0), BREAK_GLASS, GIBS_LIFE, 0 );



               /*
	MESSAGE_BEGIN( MSG_PVS, SVC_TEMPENTITY, vecSpot );
		WRITE_BYTE( TE_BREAKMODEL );
		WRITE_COORD( vecSpot.x );
		WRITE_COORD( vecSpot.y );
		WRITE_COORD( vecSpot.z );
		WRITE_COORD( pev->size.x );
		WRITE_COORD( pev->size.y );
		WRITE_COORD( pev->size.z );
		WRITE_COORD( 0 ); 
		WRITE_COORD( 0 );
		WRITE_COORD( 0 );
		WRITE_BYTE( 10 ); 
		WRITE_SHORT( m_idShard );
		WRITE_BYTE( 0 );
		WRITE_BYTE( GIBS_LIFE );
		WRITE_BYTE( BREAK_GLASS );
	MESSAGE_END();*/
}

void CFuncLight :: Flicker( void )
{
	if( m_iState == STATE_TURN_ON )
	{
		LIGHT_STYLE( m_iStyle, "k" );
		UTIL_FireTargets( pev->target, this, this, USE_ON );
		m_iState = STATE_ON;
		DontThink();
		return;
	}

	if( m_iFlickerMode == 1 )
	{
		pev->frame = 1;
		LIGHT_STYLE( m_iStyle, "a" );
		SetThink( NULL );
		return;
	}

	if( m_iFlickerMode == 2 )
	{
		switch( RANDOM_LONG( 0, 3 ))
		{
		case 0:
			LIGHT_STYLE( m_iStyle, "abcaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa" );
			break;
		case 1:
			LIGHT_STYLE( m_iStyle, "acaaabaaaaaaaaaaaaaaaaaaaaaaaaaaa" );
			break;
		case 2:
			LIGHT_STYLE( m_iStyle, "aaafbaaaaaaaaaaaaaaaaaaaaaaaaaaaa" );
			break;
		case 3:
			LIGHT_STYLE( m_iStyle, "aaaaaaaaaaaaagaaaaaaaaacaaaacaaaa" );
			break;
		}		

		m_flNextFlickerTime = RANDOM_FLOAT( 0.5f, 10.0f );
                   	UTIL_Sparks( m_vecLastDmgPoint );

		switch( RANDOM_LONG( 0, 2 ))
		{
		case 0:
			EMIT_SOUND( edict(), CHAN_VOICE, "buttons/spark1.wav", 0.4, ATTN_IDLE );
			break;
		case 1:
			EMIT_SOUND( edict(), CHAN_VOICE, "buttons/spark2.wav", 0.3, ATTN_IDLE );
			break;
		case 2:
			EMIT_SOUND( edict(), CHAN_VOICE, "buttons/spark3.wav", 0.35, ATTN_IDLE );
			break;
		}

		if( m_flNextFlickerTime > 6.5f )
			m_iFlickerMode = 1; // stop sparking
	}

	SetNextThink( m_flNextFlickerTime ); 
}