/***
*
*	Copyright (c) 1996-2002, Valve LLC. All rights reserved.
*	
*	This product contains software technology licensed from Id 
*	Software, Inc. ("Id Technology").  Id Technology (c) 1996 Id Software, Inc. 
*	All Rights Reserved.
*
*   Use, distribution, and modification of this source code and/or resulting
*   object code is restricted to non-commercial enhancements to products from
*   Valve LLC.  All other use, distribution, or modification is prohibited
*   without written permission from Valve LLC.
*
****/
#include "extdll.h"
#include "util.h"
#include "cbase.h"
#include "monsters.h"
#include "customentity.h"
#include "ent/effects.h"
#include "weapons\weapons.h"
#include "player.h"
#include "decals.h"
#include "func_break.h"
#include "shake.h"
#include "studio.h"
#include "gamerules.h"
#include "trains.h"

#define SF_GIBSHOOTER_REPEATABLE	1 // allows a gibshooter to be refired
#define SF_FUNNEL_REVERSE		1 // funnel effect repels particles instead of attracting them.

// =================== INFO_TARGET ==============================================

#define SF_TARGET_HACK_VISIBLE	BIT( 0 )

class CInfoTarget : public CPointEntity
{
	DECLARE_CLASS( CInfoTarget, CPointEntity );
public:
	void Spawn( void );
	void Precache( void );
};

LINK_ENTITY_TO_CLASS( info_target, CInfoTarget );

void CInfoTarget :: Precache( void )
{
	if( FBitSet( pev->spawnflags, SF_TARGET_HACK_VISIBLE ))
		PRECACHE_MODEL( "sprites/null.spr" );
}

void CInfoTarget :: Spawn( void )
{
	Precache();

	pev->solid = SOLID_NOT;
	SetBits( m_iFlags, MF_POINTENTITY );

	if( FBitSet( pev->spawnflags, SF_TARGET_HACK_VISIBLE ))
	{
		SET_MODEL( edict(), "sprites/null.spr" );
		UTIL_SetSize( pev, g_vecZero, g_vecZero );
	}
}

class CBubbling : public CBaseEntity
{
	DECLARE_CLASS( CBubbling, CBaseEntity );
public:
	void	Spawn( void );
	void	Precache( void );
	void	KeyValue( KeyValueData *pkvd );
	
	void	FizzThink( void );
	void	Use( CBaseEntity *pActivator, CBaseEntity *pCaller, USE_TYPE useType, float value );

	DECLARE_DATADESC();

	virtual int ObjectCaps( void ) { return BaseClass::ObjectCaps() & ~FCAP_ACROSS_TRANSITION; }

	int	m_density;
	int	m_frequency;
	int	m_bubbleModel;
	int	m_state;
};

LINK_ENTITY_TO_CLASS( env_bubbles, CBubbling );

BEGIN_DATADESC( CBubbling )
	DEFINE_KEYFIELD( m_density, FIELD_INTEGER, "density" ),
	DEFINE_KEYFIELD( m_frequency, FIELD_INTEGER, "frequency" ),
	DEFINE_FIELD( m_state, FIELD_INTEGER ),
	DEFINE_FUNCTION( FizzThink ),
END_DATADESC()


#define SF_BUBBLES_STARTOFF		0x0001

void CBubbling::Spawn( void )
{
	Precache( );
	SET_MODEL( ENT(pev), STRING(pev->model) );		// Set size

	pev->solid = SOLID_NOT;							// Remove model & collisions
	pev->renderamt = 0;								// The engine won't draw this model if this is set to 0 and blending is on
	pev->rendermode = kRenderTransTexture;
	int speed = pev->speed > 0 ? pev->speed : -pev->speed;

	// HACKHACK!!! - Speed in rendercolor
	pev->rendercolor.x = speed >> 8;
	pev->rendercolor.y = speed & 255;
	pev->rendercolor.z = (pev->speed < 0) ? 1 : 0;


	if ( !(pev->spawnflags & SF_BUBBLES_STARTOFF) )
	{
		SetThink( FizzThink );
		pev->nextthink = gpGlobals->time + 2.0;
		m_state = 1;
	}
	else 
		m_state = 0;
}

void CBubbling::Precache( void )
{
	m_bubbleModel = PRECACHE_MODEL("sprites/bubble.spr");			// Precache bubble sprite
}


void CBubbling::Use( CBaseEntity *pActivator, CBaseEntity *pCaller, USE_TYPE useType, float value )
{
	if ( ShouldToggle( useType, m_state ) )
		m_state = !m_state;

	if ( m_state )
	{
		SetThink( FizzThink );
		pev->nextthink = gpGlobals->time + 0.1;
	}
	else
	{
		SetThink( NULL );
		pev->nextthink = 0;
	}
}


void CBubbling::KeyValue( KeyValueData *pkvd )
{
	if (FStrEq(pkvd->szKeyName, "density"))
	{
		m_density = atoi(pkvd->szValue);
		pkvd->fHandled = TRUE;
	}
	else if (FStrEq(pkvd->szKeyName, "frequency"))
	{
		m_frequency = atoi(pkvd->szValue);
		pkvd->fHandled = TRUE;
	}
	else if (FStrEq(pkvd->szKeyName, "current"))
	{
		pev->speed = atoi(pkvd->szValue);
		pkvd->fHandled = TRUE;
	}
	else
		BaseClass::KeyValue( pkvd );
}


void CBubbling::FizzThink( void )
{
	MESSAGE_BEGIN( MSG_PAS, SVC_TEMPENTITY, VecBModelOrigin(pev) );
		WRITE_BYTE( TE_FIZZ );
		WRITE_SHORT( (short)ENTINDEX( edict() ) );
		WRITE_SHORT( (short)m_bubbleModel );
		WRITE_BYTE( m_density );
	MESSAGE_END();

	if ( m_frequency > 19 )
		pev->nextthink = gpGlobals->time + 0.5;
	else
		pev->nextthink = gpGlobals->time + 2.5 - (0.1 * m_frequency);
}



class CGibShooter : public CBaseDelay
{
	DECLARE_CLASS( CGibShooter, CBaseDelay );
public:
	void Spawn( void );
	void Precache( void );
	void KeyValue( KeyValueData *pkvd );
	void ShootThink( void );
	void Use( CBaseEntity *pActivator, CBaseEntity *pCaller, USE_TYPE useType, float value );
	virtual int ObjectCaps( void ) { return (BaseClass::ObjectCaps() & (~FCAP_ACROSS_TRANSITION)) | FCAP_SET_MOVEDIR; };

	virtual CBaseEntity *CreateGib( void );

	DECLARE_DATADESC();

	int	m_iGibs;
	int	m_iGibCapacity;
	int	m_iGibMaterial;
	int	m_iGibModelIndex;
	float	m_flGibVelocity;
	float	m_flVariance;
	float	m_flGibLife;
	string_t	m_iszTargetname;
	string_t	m_iszSpawnTarget;
	int	m_iBloodColor;
};

BEGIN_DATADESC( CGibShooter )
	DEFINE_KEYFIELD( m_iGibs, FIELD_INTEGER, "m_iGibs" ),
	DEFINE_KEYFIELD( m_iGibCapacity, FIELD_INTEGER, "m_iGibs" ),
	DEFINE_FIELD( m_iGibMaterial, FIELD_INTEGER ),
	DEFINE_FIELD( m_iGibModelIndex, FIELD_INTEGER ),
	DEFINE_KEYFIELD( m_flGibVelocity, FIELD_FLOAT, "m_flVelocity" ),
	DEFINE_KEYFIELD( m_flVariance, FIELD_FLOAT, "m_flVariance" ),
	DEFINE_KEYFIELD( m_flGibLife, FIELD_FLOAT, "m_flGibLife" ),
	DEFINE_KEYFIELD( m_iszTargetname, FIELD_STRING, "m_iszTargetName" ),
	DEFINE_KEYFIELD( m_iszSpawnTarget, FIELD_STRING, "m_iszSpawnTarget" ),
	DEFINE_KEYFIELD( m_iBloodColor, FIELD_INTEGER, "m_iBloodColor" ),
	DEFINE_FUNCTION( ShootThink ),
END_DATADESC()

LINK_ENTITY_TO_CLASS( gibshooter, CGibShooter );

void CGibShooter :: Precache ( void )
{
	if ( g_Language == LANGUAGE_GERMAN )
	{
		m_iGibModelIndex = PRECACHE_MODEL ("models/gibs/germanygibs.mdl");
	}
	else if (m_iBloodColor == BLOOD_COLOR_YELLOW)
	{
		m_iGibModelIndex = PRECACHE_MODEL ("models/gibs/agibs.mdl");
	}
	else
	{
		m_iGibModelIndex = PRECACHE_MODEL ("models/gibs/hgibs.mdl");
	}
}

void CGibShooter::KeyValue( KeyValueData *pkvd )
{
	if( FStrEq( pkvd->szKeyName, "m_iGibs" ))
	{
		m_iGibs = m_iGibCapacity = Q_atoi( pkvd->szValue );
		pkvd->fHandled = TRUE;
	}
	else if( FStrEq( pkvd->szKeyName, "m_flVelocity" ))
	{
		m_flGibVelocity = Q_atof( pkvd->szValue );
		pkvd->fHandled = TRUE;
	}
	else if( FStrEq( pkvd->szKeyName, "m_flVariance" ))
	{
		m_flVariance = Q_atof( pkvd->szValue );
		pkvd->fHandled = TRUE;
	}
	else if( FStrEq( pkvd->szKeyName, "m_flGibLife" ))
	{
		m_flGibLife = Q_atof( pkvd->szValue );
		pkvd->fHandled = TRUE;
	}
	else if (FStrEq(pkvd->szKeyName, "m_iszTargetName"))
	{
		m_iszTargetname = ALLOC_STRING(pkvd->szValue);
		pkvd->fHandled = TRUE;
	}
	else if (FStrEq(pkvd->szKeyName, "m_iszSpawnTarget"))
	{
		m_iszSpawnTarget = ALLOC_STRING(pkvd->szValue);
		pkvd->fHandled = TRUE;
	}
	else if (FStrEq(pkvd->szKeyName, "m_iBloodColor"))
	{
		m_iBloodColor = Q_atoi(pkvd->szValue);
		pkvd->fHandled = TRUE;
	}
	else
	{
		CBaseDelay::KeyValue( pkvd );
	}
}

void CGibShooter::Use( CBaseEntity *pActivator, CBaseEntity *pCaller, USE_TYPE useType, float value )
{
	m_hActivator = pActivator;
	SetThink( ShootThink );
	SetNextThink( 0 );
}

void CGibShooter::Spawn( void )
{
	Precache();

	pev->solid = SOLID_NOT;
	pev->effects |= EF_NODRAW;

	// is mapper forgot set angles?
	if( pev->movedir == g_vecZero )
		pev->movedir = Vector( 1, 0, 0 );

	if ( m_flDelay == 0 )
	{
		m_flDelay = 0.1;
	}

	if ( m_flGibLife == 0 )
	{
		m_flGibLife = 25;
	}

	pev->body = MODEL_FRAMES( m_iGibModelIndex );
}


CBaseEntity *CGibShooter :: CreateGib( void )
{
	if ( CVAR_GET_FLOAT("violence_hgibs") == 0 )
		return NULL;

	CGib *pGib = GetClassPtr( (CGib *)NULL );

	if (m_iBloodColor == BLOOD_COLOR_YELLOW)
	{
		pGib->Spawn( "models/gibs/agibs.mdl" );
		pGib->m_bloodColor = BLOOD_COLOR_YELLOW;
	}
	else if (m_iBloodColor)
	{
		pGib->Spawn( "models/gibs/hgibs.mdl" );
		pGib->m_bloodColor = m_iBloodColor;
	}
	else
	{
		pGib->Spawn( "models/gibs/hgibs.mdl" );
		pGib->m_bloodColor = BLOOD_COLOR_RED;
	}

	if ( pev->body <= 1 )
	{
		ALERT ( at_aiconsole, "GibShooter Body is <= 1!\n" );
	}

	pGib->pev->body = RANDOM_LONG ( 1, pev->body - 1 );// avoid throwing random amounts of the 0th gib. (skull).

	return pGib;
}


void CGibShooter :: ShootThink ( void )
{
	SetNextThink( m_flDelay );

	Vector vecShootDir = EntityToWorldTransform().VectorRotate( pev->movedir );
	UTIL_MakeVectors( GetAbsAngles( )); // g-cont. was missed into original game

	vecShootDir = vecShootDir + gpGlobals->v_right * RANDOM_FLOAT( -1, 1) * m_flVariance;
	vecShootDir = vecShootDir + gpGlobals->v_forward * RANDOM_FLOAT( -1, 1) * m_flVariance;
	vecShootDir = vecShootDir + gpGlobals->v_up * RANDOM_FLOAT( -1, 1) * m_flVariance;

	vecShootDir = vecShootDir.Normalize();
	CBaseEntity *pShot = CreateGib();
	
	if( pShot )
	{
		pShot->SetAbsOrigin( GetAbsOrigin( ));
		pShot->SetAbsVelocity( vecShootDir * m_flGibVelocity );

		if( pShot->m_iActorType == ACTOR_DYNAMIC )
			WorldPhysic->AddForce( pShot, vecShootDir * m_flGibVelocity * ( 1.0f / gpGlobals->frametime ) * 0.1f );

		// custom particles already set this
		if( FClassnameIs( pShot->pev, "gib" ))
		{
			CGib *pGib = (CGib *)pShot;

			Vector vecAvelocity = pGib->GetLocalAvelocity();	
			vecAvelocity.x = RANDOM_FLOAT( 100, 200 );
			vecAvelocity.y = RANDOM_FLOAT( 100, 300 );
			pGib->SetLocalAvelocity( vecAvelocity );

			float thinkTime = pGib->pev->nextthink - gpGlobals->time;
			pGib->m_lifeTime = (m_flGibLife * RANDOM_FLOAT( 0.95, 1.05 )); // +/- 5%

			if( pGib->m_lifeTime < thinkTime )
			{
				pGib->SetNextThink( pGib->m_lifeTime );
				pGib->m_lifeTime = 0;
			}
                    }

		pShot->pev->targetname = m_iszTargetname;

		if( m_iszSpawnTarget )
			UTIL_FireTargets( m_iszSpawnTarget, pShot, this, USE_TOGGLE, 0 );		
	}

	if( --m_iGibs <= 0 )
	{
		if( FBitSet( pev->spawnflags, SF_GIBSHOOTER_REPEATABLE ))
		{
			m_iGibs = m_iGibCapacity;
			SetThink ( NULL );
			SetNextThink( 0 );
		}
		else
		{
			SetThink ( SUB_Remove );
			SetNextThink( 0 );
		}
	}
}


class CEnvShooter : public CGibShooter
{
	DECLARE_CLASS( CEnvShooter, CGibShooter );

	void Spawn( void );
	void Precache( void );
	void KeyValue( KeyValueData *pkvd );

	DECLARE_DATADESC();

	virtual CBaseEntity *CreateGib( void );

	string_t m_iszTouch;
	string_t m_iszTouchOther;
	int m_iPhysics;
	float m_fFriction;
	Vector m_vecSize;
};

BEGIN_DATADESC( CEnvShooter )
	DEFINE_KEYFIELD( m_iszTouch, FIELD_STRING, "m_iszTouch" ),
	DEFINE_KEYFIELD( m_iszTouchOther, FIELD_STRING, "m_iszTouchOther" ),
	DEFINE_KEYFIELD( m_iPhysics, FIELD_INTEGER, "m_iPhysics" ),
	DEFINE_KEYFIELD( m_fFriction, FIELD_FLOAT, "m_fFriction" ),
	DEFINE_KEYFIELD( m_vecSize, FIELD_VECTOR, "m_vecSize" ),
END_DATADESC()

LINK_ENTITY_TO_CLASS( env_shooter, CEnvShooter );

void CEnvShooter :: KeyValue( KeyValueData *pkvd )
{
	if (FStrEq(pkvd->szKeyName, "shootmodel"))
	{
		pev->model = ALLOC_STRING(pkvd->szValue);
		pkvd->fHandled = TRUE;
	}
	else if (FStrEq(pkvd->szKeyName, "shootsounds"))
	{
		int iNoise = Q_atoi(pkvd->szValue);
		pkvd->fHandled = TRUE;
		switch( iNoise )
		{
		case 0:
			m_iGibMaterial = matGlass;
			break;
		case 1:
			m_iGibMaterial = matWood;
			break;
		case 2:
			m_iGibMaterial = matMetal;
			break;
		case 3:
			m_iGibMaterial = matFlesh;
			break;
		case 4:
			m_iGibMaterial = matRocks;
			break;
		default:
		case -1:
			m_iGibMaterial = matNone;
			break;
		}
	}
	else if (FStrEq(pkvd->szKeyName, "m_iszTouch"))
	{
		m_iszTouch = ALLOC_STRING(pkvd->szValue);
		pkvd->fHandled = TRUE;
	}
	else if (FStrEq(pkvd->szKeyName, "m_iszTouchOther"))
	{
		m_iszTouchOther = ALLOC_STRING(pkvd->szValue);
		pkvd->fHandled = TRUE;
	}
	else if (FStrEq(pkvd->szKeyName, "m_iPhysics"))
	{
		m_iPhysics = Q_atoi(pkvd->szValue);
		pkvd->fHandled = TRUE;
	}
	else if (FStrEq(pkvd->szKeyName, "m_fFriction"))
	{
		m_fFriction = Q_atof(pkvd->szValue);
		pkvd->fHandled = TRUE;
	}
	else if (FStrEq(pkvd->szKeyName, "m_vecSize"))
	{
		UTIL_StringToVector( m_vecSize, pkvd->szValue);
		m_vecSize = m_vecSize / 2;
		pkvd->fHandled = TRUE;
	}
	else
	{
		CGibShooter::KeyValue( pkvd );
	}
}

void CEnvShooter :: Precache ( void )
{
	m_iGibModelIndex = PRECACHE_MODEL( GetModel() );

	CBreakable::MaterialSoundPrecache( (Materials)m_iGibMaterial );
}

void CEnvShooter::Spawn( void )
{
	int iBody = pev->body;
	BaseClass::Spawn();
	pev->body = iBody;
}

CBaseEntity *CEnvShooter :: CreateGib( void )
{
	if( m_iPhysics <= 1 )
	{
		// normal gib or sticky gib
		CGib *pGib = GetClassPtr( (CGib *)NULL );

		pGib->Spawn( GetModel() );

		if( m_iPhysics == 1 )
		{
			// sticky gib
			pGib->pev->movetype = MOVETYPE_TOSS;
			pGib->pev->solid = SOLID_BBOX;
			pGib->SetTouch( &CGib :: StickyGibTouch );
		}

		if( m_iBloodColor )
			pGib->m_bloodColor = m_iBloodColor;
		else pGib->m_bloodColor = DONT_BLEED;
		pGib->m_material = m_iGibMaterial;

		pGib->pev->rendermode = pev->rendermode;
		pGib->pev->renderamt = pev->renderamt;
		pGib->pev->rendercolor = pev->rendercolor;
		pGib->pev->renderfx = pev->renderfx;
		pGib->pev->effects = pev->effects & ~EF_NODRAW;
		pGib->pev->scale = pev->scale;
		pGib->pev->skin = pev->skin;

		// g-cont. random body select
		if( pGib->pev->body == -1 )
		{
			int numBodies = MODEL_FRAMES( pGib->pev->modelindex );
			pGib->pev->body = RANDOM_LONG( 0, numBodies - 1 );
		}
		else if( pev->body > 0 )
			pGib->pev->body = RANDOM_LONG( 0, pev->body - 1 );

		// g-cont. random skin select
		if( pGib->pev->skin == -1 )
		{
			studiohdr_t *pstudiohdr = (studiohdr_t *)(GET_MODEL_PTR( ENT(pGib->pev) ));

			// NOTE: this code works fine only under XashXT because GoldSource doesn't merge texture and model files
			// into one model_t slot. So this will not working with models which keep textures seperate
			if( pstudiohdr && pstudiohdr->numskinfamilies > 0 )
			{
				pGib->pev->skin = RANDOM_LONG( 0, pstudiohdr->numskinfamilies - 1 );
			}
			else pGib->pev->skin = 0; // reset to default
		}

		return pGib;
	}
          
	// special shot
	CShot *pShot = GetClassPtr( (CShot*)NULL );
	pShot->pev->classname = MAKE_STRING( "shot" );
	pShot->pev->solid = SOLID_SLIDEBOX;
	pShot->SetLocalAvelocity( GetLocalAvelocity( ));
	SET_MODEL( ENT( pShot->pev ), STRING( pev->model ));
	UTIL_SetSize( pShot->pev, -m_vecSize, m_vecSize );
	pShot->pev->renderamt = pev->renderamt;
	pShot->pev->rendermode = pev->rendermode;
	pShot->pev->rendercolor = pev->rendercolor;
	pShot->pev->renderfx = pev->renderfx;
	pShot->pev->netname = m_iszTouch;
	pShot->pev->message = m_iszTouchOther;
	pShot->pev->effects = pev->effects & ~EF_NODRAW;
	pShot->pev->skin = pev->skin;
	pShot->pev->body = pev->body;
	pShot->pev->scale = pev->scale;
	pShot->pev->frame = pev->frame;
	pShot->pev->framerate = pev->framerate;
	pShot->pev->friction = m_fFriction;

	// g-cont. random body select
	if( pShot->pev->body == -1 )
	{
		int numBodies = MODEL_FRAMES( pShot->pev->modelindex );
		pShot->pev->body = RANDOM_LONG( 0, numBodies - 1 );
	}
	else if( pev->body > 0 )
		pShot->pev->body = RANDOM_LONG( 0, pev->body - 1 );

	// g-cont. random skin select
	if( pShot->pev->skin == -1 )
	{
		studiohdr_t *pstudiohdr = (studiohdr_t *)(GET_MODEL_PTR( ENT(pShot->pev) ));

		// NOTE: this code works fine only under XashXT because GoldSource doesn't merge texture and model files
		// into one model_t slot. So this will not working with models which keep textures seperate
		if( pstudiohdr && pstudiohdr->numskinfamilies > 0 )
		{
			pShot->pev->skin = RANDOM_LONG( 0, pstudiohdr->numskinfamilies - 1 );
		}
		else pShot->pev->skin = 0; // reset to default
	}

	switch( m_iPhysics )
	{
	case 2:	
		pShot->pev->movetype = MOVETYPE_NOCLIP;
		pShot->pev->solid = SOLID_NOT;
		break;
	case 3:
		pShot->pev->movetype = MOVETYPE_FLYMISSILE;
		break;
	case 4:
		pShot->pev->movetype = MOVETYPE_BOUNCEMISSILE;
		break;
	case 5:
		pShot->pev->movetype = MOVETYPE_TOSS;
		break;
	case 6:
		// FIXME: tune NxGroupMask for each body to avoid collision with this particles
		// because so small bodies can be pushed through floor
		pShot->pev->solid = SOLID_NOT;
		if( WorldPhysic->Initialized( ))
			pShot->pev->movetype = MOVETYPE_PHYSIC;
		pShot->SetAbsOrigin( GetAbsOrigin( ));
		pShot->m_pUserData = WorldPhysic->CreateBodyFromEntity( pShot );
		break;
	default:
		pShot->pev->movetype = MOVETYPE_BOUNCE;
		break;
	}

	if( pShot->pev->framerate && UTIL_GetModelType( pShot->pev->modelindex ) == mod_sprite )
	{
		pShot->m_maxFrame = (float)MODEL_FRAMES( pShot->pev->modelindex ) - 1;
		if( pShot->m_maxFrame > 1.0f )
		{
			if( m_flGibLife )
			{
				pShot->pev->dmgtime = gpGlobals->time + m_flGibLife;
				pShot->SetThink( &CSprite::AnimateUntilDead );
			}
			else
			{
				pShot->SetThink( &CSprite::AnimateThink );
			}

			pShot->SetNextThink( 0 );
			pShot->m_lastTime = gpGlobals->time;
			return pShot;
		}
	}

	// if it's not animating
	if( m_flGibLife )
	{
		pShot->SetThink( SUB_Remove );
		pShot->SetNextThink( m_flGibLife );
	}

	return pShot;
}

class CTestEffect : public CBaseDelay
{
	DECLARE_CLASS( CTestEffect, CBaseDelay );
public:
	void	Spawn( void );
	void	Precache( void );

	void	TestThink( void );
	void	Use( CBaseEntity *pActivator, CBaseEntity *pCaller, USE_TYPE useType, float value );

	DECLARE_DATADESC();

	int	m_iLoop;
	int	m_iBeam;
	CBeam	*m_pBeam[24];
	float	m_flBeamTime[24];
	float	m_flStartTime;
};

LINK_ENTITY_TO_CLASS( test_effect, CTestEffect );

BEGIN_DATADESC( CTestEffect )
	DEFINE_FUNCTION( TestThink ),
END_DATADESC()

void CTestEffect::Spawn( void )
{
	Precache( );
}

void CTestEffect::Precache( void )
{
	PRECACHE_MODEL( "sprites/beam/lgtning.spr" );
}

void CTestEffect::TestThink( void )
{
	int i;
	float t = (gpGlobals->time - m_flStartTime);

	if (m_iBeam < 24)
	{
		CBeam *pbeam = CBeam::BeamCreate( "sprites/beam/lgtning.spr", 100 );

		TraceResult		tr;

		Vector vecSrc = GetAbsOrigin();
		Vector vecDir = Vector( RANDOM_FLOAT( -1.0, 1.0 ), RANDOM_FLOAT( -1.0, 1.0 ),RANDOM_FLOAT( -1.0, 1.0 ) );
		vecDir = vecDir.Normalize();
		UTIL_TraceLine( vecSrc, vecSrc + vecDir * 128, ignore_monsters, ENT(pev), &tr);

		pbeam->PointsInit( vecSrc, tr.vecEndPos );
		// pbeam->SetColor( 80, 100, 255 );
		pbeam->SetColor( 255, 180, 100 );
		pbeam->SetWidth( 100 );
		pbeam->SetScrollRate( 12 );
		
		m_flBeamTime[m_iBeam] = gpGlobals->time;
		m_pBeam[m_iBeam] = pbeam;
		m_iBeam++;
	}

	if (t < 3.0)
	{
		for (i = 0; i < m_iBeam; i++)
		{
			t = (gpGlobals->time - m_flBeamTime[i]) / ( 3 + m_flStartTime - m_flBeamTime[i]);
			m_pBeam[i]->SetBrightness( 255 * t );
			// m_pBeam[i]->SetScrollRate( 20 * t );
		}
		pev->nextthink = gpGlobals->time + 0.1;
	}
	else
	{
		for (i = 0; i < m_iBeam; i++)
		{
			UTIL_Remove( m_pBeam[i] );
		}
		m_flStartTime = gpGlobals->time;
		m_iBeam = 0;
		// pev->nextthink = gpGlobals->time;
		SetThink( NULL );
	}
}


void CTestEffect::Use( CBaseEntity *pActivator, CBaseEntity *pCaller, USE_TYPE useType, float value )
{
	SetThink( TestThink );
	pev->nextthink = gpGlobals->time + 0.1;
	m_flStartTime = gpGlobals->time;
}



// Blood effects
class CBlood : public CPointEntity
{
	DECLARE_CLASS( CBlood, CPointEntity );
public:
	void	Spawn( void );
	void	Use( CBaseEntity *pActivator, CBaseEntity *pCaller, USE_TYPE useType, float value );
	virtual int ObjectCaps( void ) { return (BaseClass :: ObjectCaps() & ~FCAP_ACROSS_TRANSITION) | FCAP_SET_MOVEDIR; }
	void	KeyValue( KeyValueData *pkvd );

	inline	int Color( void ) { return pev->impulse; }
	inline	float BloodAmount( void ) { return pev->dmg; }

	inline	void SetColor( int color ) { pev->impulse = color; }
	inline	void SetBloodAmount( float amount ) { pev->dmg = amount; }
	
	Vector	Direction( void );
	Vector	BloodPosition( CBaseEntity *pActivator );
};

LINK_ENTITY_TO_CLASS( env_blood, CBlood );

#define SF_BLOOD_RANDOM		0x0001
#define SF_BLOOD_STREAM		0x0002
#define SF_BLOOD_PLAYER		0x0004
#define SF_BLOOD_DECAL		0x0008

void CBlood::Spawn( void )
{
	pev->solid	= SOLID_NOT;
	pev->movetype	= MOVETYPE_NONE;
	pev->frame	= 0;
}

void CBlood::KeyValue( KeyValueData *pkvd )
{
	if (FStrEq(pkvd->szKeyName, "color"))
	{
		int color = atoi(pkvd->szValue);
		switch( color )
		{
		case 1:
			SetColor( BLOOD_COLOR_YELLOW );
			break;
		default:
			SetColor( BLOOD_COLOR_RED );
			break;
		}

		pkvd->fHandled = TRUE;
	}
	else if (FStrEq(pkvd->szKeyName, "amount"))
	{
		SetBloodAmount( atof(pkvd->szValue) );
		pkvd->fHandled = TRUE;
	}
	else
		CPointEntity::KeyValue( pkvd );
}

Vector CBlood::Direction( void )
{
	if ( pev->spawnflags & SF_BLOOD_RANDOM )
		return UTIL_RandomBloodVector();
	
	return pev->movedir;
}

Vector CBlood::BloodPosition( CBaseEntity *pActivator )
{
	if ( pev->spawnflags & SF_BLOOD_PLAYER )
	{
		edict_t *pPlayer;

		if ( pActivator && pActivator->IsPlayer() )
		{
			pPlayer = pActivator->edict();
		}
		else
			pPlayer = INDEXENT( 1 );
		if ( pPlayer )
			return (pPlayer->v.origin + pPlayer->v.view_ofs) + Vector( RANDOM_FLOAT(-10,10), RANDOM_FLOAT(-10,10), RANDOM_FLOAT(-10,10) );
	}

	return GetAbsOrigin();
}


void CBlood::Use( CBaseEntity *pActivator, CBaseEntity *pCaller, USE_TYPE useType, float value )
{
	if ( pev->spawnflags & SF_BLOOD_STREAM )
		UTIL_BloodStream( BloodPosition(pActivator), Direction(), (Color() == BLOOD_COLOR_RED) ? 70 : Color(), BloodAmount() );
	else
		UTIL_BloodDrips( BloodPosition(pActivator), Direction(), Color(), BloodAmount() );

	if ( pev->spawnflags & SF_BLOOD_DECAL )
	{
		Vector forward = Direction();
		Vector start = BloodPosition( pActivator );
		TraceResult tr;

		UTIL_TraceLine( start, start + forward * BloodAmount() * 2, ignore_monsters, NULL, &tr );
		if ( tr.flFraction != 1.0 )
			UTIL_BloodDecalTrace( &tr, Color() );
	}
}



// Screen shake
class CShake : public CPointEntity
{
	DECLARE_CLASS( CShake, CPointEntity );
public:
	void	Spawn( void );
	void	Use( CBaseEntity *pActivator, CBaseEntity *pCaller, USE_TYPE useType, float value );
	void	KeyValue( KeyValueData *pkvd );

	inline	float Amplitude( void ) { return pev->scale; }
	inline	float Frequency( void ) { return pev->dmg_save; }
	inline	float Duration( void ) { return pev->dmg_take; }
	inline	float Radius( void ) { return pev->dmg; }

	inline	void SetAmplitude( float amplitude ) { pev->scale = amplitude; }
	inline	void SetFrequency( float frequency ) { pev->dmg_save = frequency; }
	inline	void SetDuration( float duration ) { pev->dmg_take = duration; }
	inline	void SetRadius( float radius ) { pev->dmg = radius; }
private:
};

LINK_ENTITY_TO_CLASS( env_shake, CShake );
LINK_ENTITY_TO_CLASS( hud_shake, CShake );

// pev->scale is amplitude
// pev->dmg_save is frequency
// pev->dmg_take is duration
// pev->dmg is radius
// radius of 0 means all players
// NOTE: UTIL_ScreenShake() will only shake players who are on the ground

#define SF_SHAKE_EVERYONE	0x0001		// Don't check radius
// UNDONE: These don't work yet
#define SF_SHAKE_DISRUPT	0x0002		// Disrupt controls
#define SF_SHAKE_INAIR		0x0004		// Shake players in air

void CShake::Spawn( void )
{
	pev->solid	= SOLID_NOT;
	pev->movetype	= MOVETYPE_NONE;
	pev->frame	= 0;
	
	if ( pev->spawnflags & SF_SHAKE_EVERYONE )
		pev->dmg = 0;
}


void CShake::KeyValue( KeyValueData *pkvd )
{
	if (FStrEq(pkvd->szKeyName, "amplitude"))
	{
		SetAmplitude( atof(pkvd->szValue) );
		pkvd->fHandled = TRUE;
	}
	else if (FStrEq(pkvd->szKeyName, "frequency"))
	{
		SetFrequency( atof(pkvd->szValue) );
		pkvd->fHandled = TRUE;
	}
	else if (FStrEq(pkvd->szKeyName, "duration"))
	{
		SetDuration( atof(pkvd->szValue) );
		pkvd->fHandled = TRUE;
	}
	else if (FStrEq(pkvd->szKeyName, "radius"))
	{
		SetRadius( atof(pkvd->szValue) );
		pkvd->fHandled = TRUE;
	}
	else
		CPointEntity::KeyValue( pkvd );
}


void CShake::Use( CBaseEntity *pActivator, CBaseEntity *pCaller, USE_TYPE useType, float value )
{
	UTIL_ScreenShake( GetAbsOrigin(), Amplitude(), Frequency(), Duration(), Radius(), FBitSet( pev->spawnflags, SF_SHAKE_INAIR ) ? true : false );
}


class CFade : public CPointEntity
{
	DECLARE_CLASS( CFade, CPointEntity );
public:
	void	Spawn( void );
	void	Use( CBaseEntity *pActivator, CBaseEntity *pCaller, USE_TYPE useType, float value );
	void	KeyValue( KeyValueData *pkvd );

	inline	float Duration( void ) { return pev->dmg_take; }
	inline	float HoldTime( void ) { return pev->dmg_save; }

	inline	void SetDuration( float duration ) { pev->dmg_take = duration; }
	inline	void SetHoldTime( float hold ) { pev->dmg_save = hold; }
};

LINK_ENTITY_TO_CLASS( hud_fade, CFade );
LINK_ENTITY_TO_CLASS( env_fade, CFade );

// pev->dmg_take is duration
// pev->dmg_save is hold duration
#define SF_FADE_IN			0x0001		// Fade in, not out
#define SF_FADE_MODULATE		0x0002		// Modulate, don't blend
#define SF_FADE_ONLYONE		0x0004
#define SF_FADE_PERMANENT		0x0008		// LRC - hold permanently

void CFade::Spawn( void )
{
	pev->solid	= SOLID_NOT;
	pev->movetype	= MOVETYPE_NONE;
	pev->frame	= 0;
}


void CFade::KeyValue( KeyValueData *pkvd )
{
	if (FStrEq(pkvd->szKeyName, "duration"))
	{
		SetDuration( atof(pkvd->szValue) );
		pkvd->fHandled = TRUE;
	}
	else if (FStrEq(pkvd->szKeyName, "holdtime"))
	{
		SetHoldTime( atof(pkvd->szValue) );
		pkvd->fHandled = TRUE;
	}
	else
		CPointEntity::KeyValue( pkvd );
}


void CFade::Use( CBaseEntity *pActivator, CBaseEntity *pCaller, USE_TYPE useType, float value )
{
	int fadeFlags = 0;
	
	if ( !(pev->spawnflags & SF_FADE_IN) )
		fadeFlags |= FFADE_OUT;

	if ( pev->spawnflags & SF_FADE_MODULATE )
		fadeFlags |= FFADE_MODULATE;

	if ( pev->spawnflags & SF_FADE_PERMANENT )	//LRC
		fadeFlags |= FFADE_STAYOUT;				//LRC

	if ( pev->spawnflags & SF_FADE_ONLYONE )
	{
		if ( pActivator->IsNetClient() )
		{
			UTIL_ScreenFade( pActivator, pev->rendercolor, Duration(), HoldTime(), pev->renderamt, fadeFlags );
		}
	}
	else
	{
		UTIL_ScreenFadeAll( pev->rendercolor, Duration(), HoldTime(), pev->renderamt, fadeFlags );
	}
	SUB_UseTargets( this, USE_TOGGLE, 0 );
}


class CMessage : public CPointEntity
{
	DECLARE_CLASS( CMessage, CPointEntity );
public:
	void	Spawn( void );
	void	Precache( void );
	void	Use( CBaseEntity *pActivator, CBaseEntity *pCaller, USE_TYPE useType, float value );
	void	KeyValue( KeyValueData *pkvd );
};

LINK_ENTITY_TO_CLASS( env_message, CMessage );
LINK_ENTITY_TO_CLASS( hud_message, CMessage );

void CMessage::Spawn( void )
{
	Precache();

	pev->solid = SOLID_NOT;
	pev->movetype = MOVETYPE_NONE;

	switch( pev->impulse )
	{
	case 1: // Medium radius
		pev->speed = ATTN_STATIC;
		break;
	
	case 2:	// Large radius
		pev->speed = ATTN_NORM;
		break;

	case 3:	//EVERYWHERE
		pev->speed = ATTN_NONE;
		break;
	
	default:
	case 0: // Small radius
		pev->speed = ATTN_IDLE;
		break;
	}
	pev->impulse = 0;

	// No volume, use normal
	if ( pev->scale <= 0 )
		pev->scale = 1.0;
}


void CMessage::Precache( void )
{
	if ( pev->noise )
		PRECACHE_SOUND( (char *)STRING(pev->noise) );
}

void CMessage::KeyValue( KeyValueData *pkvd )
{
	if (FStrEq(pkvd->szKeyName, "messagesound"))
	{
		pev->noise = ALLOC_STRING(pkvd->szValue);
		pkvd->fHandled = TRUE;
	}
	else if (FStrEq(pkvd->szKeyName, "messagevolume"))
	{
		pev->scale = atof(pkvd->szValue) * 0.1;
		pkvd->fHandled = TRUE;
	}
	else if (FStrEq(pkvd->szKeyName, "messageattenuation"))
	{
		pev->impulse = atoi(pkvd->szValue);
		pkvd->fHandled = TRUE;
	}
	else
		BaseClass::KeyValue( pkvd );
}

void CMessage::Use( CBaseEntity *pActivator, CBaseEntity *pCaller, USE_TYPE useType, float value )
{
	CBaseEntity *pPlayer = NULL;

	if ( pev->spawnflags & SF_MESSAGE_ALL )
		UTIL_ShowMessageAll( STRING(pev->message) );
	else
	{
		if ( pActivator && pActivator->IsPlayer() )
			pPlayer = pActivator;
		else
		{
			pPlayer = CBaseEntity::Instance( INDEXENT( 1 ) );
		}
		if ( pPlayer )
			UTIL_ShowMessage( STRING(pev->message), pPlayer );
	}
	if ( pev->noise )
	{
		EMIT_SOUND( edict(), CHAN_BODY, STRING(pev->noise), pev->scale, pev->speed );
	}
	if ( pev->spawnflags & SF_MESSAGE_ONCE )
		UTIL_Remove( this );

	SUB_UseTargets( this, USE_TOGGLE, 0 );
}



//=========================================================
// FunnelEffect
//=========================================================
class CEnvFunnel : public CBaseDelay
{
	DECLARE_CLASS( CEnvFunnel, CBaseDelay );
public:
	void	Spawn( void );
	void	Precache( void );
	void	Use( CBaseEntity *pActivator, CBaseEntity *pCaller, USE_TYPE useType, float value );

	int	m_iSprite;	// Don't save, precache
};

void CEnvFunnel :: Precache ( void )
{
	if (pev->netname)
		m_iSprite = PRECACHE_MODEL ( (char*)STRING(pev->netname) );
	else
		m_iSprite = PRECACHE_MODEL ( "sprites/flare/flare6.spr" );
}

LINK_ENTITY_TO_CLASS( env_funnel, CEnvFunnel );

void CEnvFunnel::Use( CBaseEntity *pActivator, CBaseEntity *pCaller, USE_TYPE useType, float value )
{
	Vector absOrigin = GetAbsOrigin();

	MESSAGE_BEGIN( MSG_BROADCAST, SVC_TEMPENTITY );
		WRITE_BYTE( TE_LARGEFUNNEL );
		WRITE_COORD( absOrigin.x );
		WRITE_COORD( absOrigin.y );
		WRITE_COORD( absOrigin.z );
		WRITE_SHORT( m_iSprite );

		if ( pev->spawnflags & SF_FUNNEL_REVERSE )// funnel flows in reverse?
		{
			WRITE_SHORT( 1 );
		}
		else
		{
			WRITE_SHORT( 0 );
		}


	MESSAGE_END();

	SetThink( SUB_Remove );
	pev->nextthink = gpGlobals->time;
}

void CEnvFunnel::Spawn( void )
{
	Precache();
	pev->solid = SOLID_NOT;
	pev->effects |= EF_NODRAW;
}

//=========================================================
// Beverage Dispenser
// overloaded pev->frags, is now a flag for whether or not a can is stuck in the dispenser. 
// overloaded pev->health, is now how many cans remain in the machine.
//=========================================================
class CEnvBeverage : public CBaseDelay
{
	DECLARE_CLASS( CEnvBeverage, CBaseDelay );
public:
	void	Spawn( void );
	void	Precache( void );
	void	Use( CBaseEntity *pActivator, CBaseEntity *pCaller, USE_TYPE useType, float value );
};

void CEnvBeverage :: Precache ( void )
{
	PRECACHE_MODEL( "models/items/drink.mdl" );
	PRECACHE_SOUND( "weapons/g_bounce3.wav" );
}

LINK_ENTITY_TO_CLASS( env_beverage, CEnvBeverage );

void CEnvBeverage::Use( CBaseEntity *pActivator, CBaseEntity *pCaller, USE_TYPE useType, float value )
{
	if ( pev->frags != 0 || pev->health <= 0 )
	{
		// no more cans while one is waiting in the dispenser, or if I'm out of cans.
		return;
	}

	CBaseEntity *pCan = CBaseEntity::Create( "item_sodacan", GetAbsOrigin(), GetAbsAngles(), edict() );	

	if ( pev->skin == 6 )
	{
		// random
		pCan->pev->skin = RANDOM_LONG( 0, 5 );
	}
	else
	{
		pCan->pev->skin = pev->skin;
	}

	pev->frags = 1;
	pev->health--;

	//SetThink (SUB_Remove);
	//pev->nextthink = gpGlobals->time;
}

void CEnvBeverage::Spawn( void )
{
	Precache();
	pev->solid = SOLID_NOT;
	pev->effects |= EF_NODRAW;
	pev->frags = 0;

	if ( pev->health == 0 )
	{
		pev->health = 10;
	}
}

//=========================================================
// Soda can
//=========================================================
class CItemSoda : public CBaseEntity
{
	DECLARE_CLASS( CItemSoda, CBaseEntity );
public:
	void	Spawn( void );
	void	Precache( void );
	void	CanThink ( void );
	void	CanTouch ( CBaseEntity *pOther );
virtual int ObjectCaps( void ) { return (BaseClass :: ObjectCaps()) | FCAP_HOLDABLE_ITEM; }
	DECLARE_DATADESC();
};

BEGIN_DATADESC( CItemSoda )
	DEFINE_FUNCTION( CanThink ),
	DEFINE_FUNCTION( CanTouch ),
END_DATADESC()

void CItemSoda :: Precache ( void )
{
	PRECACHE_MODEL( "models/items/drink.mdl" );
}

LINK_ENTITY_TO_CLASS( item_sodacan, CItemSoda );

void CItemSoda::Spawn( void )
{
	Precache();
	pev->solid = SOLID_CUSTOM;

	if( WorldPhysic->Initialized( ))
		pev->movetype = MOVETYPE_PHYSIC;
	else pev->movetype = MOVETYPE_TOSS;

	SET_MODEL ( ENT(pev), "models/items/drink.mdl" );
	UTIL_SetSize ( pev, Vector ( 0, 0, 0 ), Vector ( 0, 0, 0 ) );
	
	SetThink (CanThink);
	pev->nextthink = gpGlobals->time + 0.5;

	m_pUserData = WorldPhysic->CreateBodyFromEntity( this );
}

void CItemSoda::CanThink ( void )
{
	EMIT_SOUND (ENT(pev), CHAN_WEAPON, "weapons/g_bounce3.wav", 1, ATTN_NORM );

	pev->solid = SOLID_TRIGGER;
	UTIL_SetSize ( pev, Vector ( -8, -8, 0 ), Vector ( 8, 8, 8 ) );
	SetThink ( NULL );
	SetTouch ( CanTouch );
}

void CItemSoda::CanTouch ( CBaseEntity *pOther )
{
	if ( !pOther->IsPlayer() )
	{
		return;
	}

	// spoit sound here

	pOther->TakeHealth( 1, DMG_GENERIC );// a bit of health.

	if ( !FNullEnt( pev->owner ) )
	{
		// tell the machine the can was taken
		pev->owner->v.frags = 0;
	}

	pev->solid = SOLID_NOT;
	pev->movetype = MOVETYPE_NONE;
	pev->effects |= EF_NODRAW;
	SetTouch ( NULL );
	SetThink ( SUB_Remove );
	pev->nextthink = gpGlobals->time;
}

// =================== ENV_SKY ==============================================
#define SF_ENVSKY_START_OFF	BIT( 0 )

class CEnvSky : public CPointEntity
{
	DECLARE_CLASS( CEnvSky, CPointEntity );
public: 
	void Spawn( void );
	void KeyValue( KeyValueData *pkvd );
	void Use( CBaseEntity *pActivator, CBaseEntity *pCaller, USE_TYPE useType, float value );
};

LINK_ENTITY_TO_CLASS( env_sky, CEnvSky );

void CEnvSky :: KeyValue( KeyValueData *pkvd )
{
	if( FStrEq( pkvd->szKeyName, "fov" ))
	{
		pev->fuser2 = Q_atof( pkvd->szValue );
		pkvd->fHandled = TRUE;
	}
	else CPointEntity::KeyValue( pkvd );
}

void CEnvSky::Spawn( void )
{
	if( !FBitSet( pev->spawnflags, SF_ENVSKY_START_OFF ))
		pev->effects |= (EF_MERGE_VISIBILITY|EF_SKYCAMERA);

	SET_MODEL( edict(), "sprites/null.spr" );
	SetBits( m_iFlags, MF_POINTENTITY );
	RelinkEntity( FALSE );
}

void CEnvSky::Use( CBaseEntity *pActivator, CBaseEntity *pCaller, USE_TYPE useType, float value )
{
	int m_active = FBitSet( pev->effects, EF_SKYCAMERA );

	if ( !ShouldToggle( useType, m_active ))
		return;

	if ( m_active )
	{
		pev->effects &= ~(EF_MERGE_VISIBILITY|EF_SKYCAMERA);
	}
	else
	{
		pev->effects |= (EF_MERGE_VISIBILITY|EF_SKYCAMERA);
	}
}

// =================== LIGHT_PROJECTOR ==============================================

#define SF_PROJECTOR_START_ON		BIT( 0 )
#define SF_PROJECTOR_NO_ATTEN		BIT( 1 )
#define SF_PROJECTOR_ASPECT4X3	BIT( 2 )	// use 4x3 aspect instead of quad
#define SF_PROJECTOR_ASPECT3X4	BIT( 3 )	// use 3x4 aspect instead of quad
// g-cont. one bit reserved for me
#define SF_PROJECTOR_NOWORLDLIGHT	BIT( 5 )	// don't apply projection to the world brushes
#define SF_PROJECTOR_NOSHADOWS	BIT( 6 )	// ignore shadows for this light (perf)
#define SF_PROJECTOR_FLIPTEXTURE	BIT( 7 )	// mirror image from right to left side

class CEnvProjector : public CBaseDelay
{
	DECLARE_CLASS( CEnvProjector, CBaseDelay );
public: 
	void Spawn( void );
	void Precache( void );
	void KeyValue( KeyValueData *pkvd );
	void Use( CBaseEntity *pActivator, CBaseEntity *pCaller, USE_TYPE useType, float value );
	virtual STATE GetState( void ) { return FBitSet( pev->effects, EF_NODRAW ) ? STATE_OFF : STATE_ON; };
	virtual int ObjectCaps( void ) { return BaseClass :: ObjectCaps() & ~FCAP_ACROSS_TRANSITION; }
	void CineThink( void );
	void SpriteThink( void );
	void PVSThink( void );
	void UpdatePVSPoint( void );
	float Frames( void ) { return pev->frags; }

	DECLARE_DATADESC();
};

LINK_ENTITY_TO_CLASS( light_projector, CEnvProjector );

BEGIN_DATADESC( CEnvProjector )
	DEFINE_FUNCTION( CineThink ),
	DEFINE_FUNCTION( SpriteThink ),
	DEFINE_FUNCTION( PVSThink ),
END_DATADESC()

void CEnvProjector :: KeyValue( KeyValueData *pkvd )
{
	if( FStrEq( pkvd->szKeyName, "fov" ))
	{
		pev->iuser2 = Q_atoi( pkvd->szValue );
		pev->iuser2 = bound( 1, pev->iuser2, 120 );
		pkvd->fHandled = TRUE;
	}
	else if( FStrEq( pkvd->szKeyName, "radius" ))
	{
		pev->scale = Q_atof( pkvd->szValue ) * (1.0f/8.0f);
		pkvd->fHandled = TRUE;
	}
	else if( FStrEq( pkvd->szKeyName, "texture" ))
	{
		pev->message = ALLOC_STRING( pkvd->szValue );
		pkvd->fHandled = TRUE;
	}
	else if( FStrEq( pkvd->szKeyName, "falloff" ))
	{
		pev->renderamt = Q_atoi( pkvd->szValue );
		pkvd->fHandled = TRUE;
	}
	else if( FStrEq( pkvd->szKeyName, "style" ))
	{
		pev->renderfx = Q_atoi( pkvd->szValue );
		pkvd->fHandled = TRUE;
	}
	else BaseClass::KeyValue( pkvd );
}

void CEnvProjector::Precache( void )
{
	if( !FStringNull( pev->message ))
	{
		if( FBitSet( pev->iuser1, ( CF_TEXTURE|CF_SPRITE )))
		{
			pev->sequence = g_engfuncs.pfnPrecacheGeneric( STRING( pev->message ));

			if( FBitSet( pev->iuser1, CF_SPRITE ))
				pev->frags = MODEL_FRAMES( PRECACHE_MODEL( (char *)STRING( pev->message )) );	// BUGBUG: this loaded sprites twice
		}
		else if( FBitSet( pev->iuser1, CF_MOVIE ))
			pev->sequence = UTIL_PrecacheMovie( pev->message );
	}
}

void CEnvProjector::Spawn( void )
{
	pev->effects |= (EF_PROJECTED_LIGHT|EF_NODRAW);
	pev->rendermode = kRenderTransTexture;
	pev->movetype = MOVETYPE_NOCLIP;

	if( !pev->framerate )
		pev->framerate = 10; // for sprites

	if( !pev->scale )
		pev->scale = 300 * (1.0f/8.0f);

	pev->frags = pev->scale; // member radius

	SET_MODEL( edict(), "sprites/null.spr" );
	SetBits( m_iFlags, MF_POINTENTITY );
//	SetLocalAvelocity( Vector( 30, 30, 30 ));
	RelinkEntity( FALSE );

	if( FBitSet( pev->spawnflags, SF_PROJECTOR_NO_ATTEN ))
		pev->iuser1 |= CF_NOATTEN;

	if( FBitSet( pev->spawnflags, SF_PROJECTOR_ASPECT4X3 ))
		pev->iuser1 |= CF_ASPECT4X3;

	if( FBitSet( pev->spawnflags, SF_PROJECTOR_ASPECT3X4 ))
		pev->iuser1 |= CF_ASPECT3X4;

	if( FBitSet( pev->spawnflags, SF_PROJECTOR_NOWORLDLIGHT ))
		pev->iuser1 |= CF_NOWORLD_PROJECTION;

	if( FBitSet( pev->spawnflags, SF_PROJECTOR_NOSHADOWS ))
		pev->iuser1 |= CF_NOSHADOWS;

	if( FBitSet( pev->spawnflags, SF_PROJECTOR_FLIPTEXTURE ))
		pev->iuser1 |= CF_FLIPTEXTURE;

	// determine texture type by extension
	if( !FStringNull( pev->message ))
	{
		const char *ext = UTIL_FileExtension( STRING( pev->message ));

		if( !Q_stricmp( ext, "tga" ) || !Q_stricmp( ext, "mip" ))
			pev->iuser1 |= CF_TEXTURE;
		else if( !Q_stricmp( ext, "spr" ))
			pev->iuser1 |= CF_SPRITE;
		else if( !Q_stricmp( ext, "avi" ))
			pev->iuser1 |= CF_MOVIE;
	}

	Precache();

	// create second PVS point
	pev->enemy = Create( "info_target", GetAbsOrigin(), g_vecZero, edict() )->edict();
	SET_MODEL( pev->enemy, "sprites/null.spr" ); // allow to collect visibility info
	UTIL_SetSize ( VARS( pev->enemy ), Vector ( -8, -8, -8 ), Vector ( 8, 8, 8 ) );

	// enable monitor
	if( FStringNull( pev->targetname ) || FBitSet( pev->spawnflags, SF_PROJECTOR_START_ON ))
	{
		SetThink( SUB_CallUseToggle );
		SetNextThink( 0.1 );
	}
}

void CEnvProjector :: Use( CBaseEntity *pActivator, CBaseEntity *pCaller, USE_TYPE useType, float value )
{
	if( IsLockedByMaster( pActivator ))
		return;

 	if( useType == USE_SET )
 	{
 		// set radius
 		value = bound( 0.0f, value, 1.0f );
		pev->scale = pev->frags * value;
		return;
	}

	if( FBitSet( pev->effects, EF_NODRAW ))
		pev->effects &= ~EF_NODRAW;
	else pev->effects |= EF_NODRAW;
          
	if( pev->effects & EF_NODRAW )
	{
		DontThink();
		return;
	}
	else
	{
		pev->dmgtime = gpGlobals->time + 0.1f;

		// run properly method
		if( FBitSet( pev->iuser1, CF_MOVIE ))
			SetThink( CineThink );
		else if( FBitSet( pev->iuser1, CF_SPRITE ) && Frames() > 1 )
			SetThink( SpriteThink );
		else SetThink( PVSThink );

		SetNextThink( 0.1f );
	}
}

void CEnvProjector :: CineThink( void )
{
	UpdatePVSPoint();

	// update as 30 frames per second
	pev->fuser2 += CIN_FRAMETIME;
	SetNextThink( CIN_FRAMETIME );
}

void CEnvProjector :: SpriteThink( void )
{
	// animate the sprite
	pev->frame += pev->framerate * ( gpGlobals->time - pev->dmgtime );
	if( pev->frame > Frames( ))
		pev->frame = fmod( pev->frame, Frames( ));
	pev->dmgtime = gpGlobals->time;

	UpdatePVSPoint();

	// keep think at 0.1 so interpolation will working properly
	SetNextThink( 0.1f );
}

void CEnvProjector :: PVSThink( void )
{
	UpdatePVSPoint();

	SetNextThink( 0.1f );
}

void CEnvProjector :: UpdatePVSPoint( void )
{
	TraceResult tr;

	UTIL_MakeVectors( GetAbsAngles() );
	Vector vecSrc = GetAbsOrigin() + gpGlobals->v_forward * 8.0f;
	Vector vecEnd = vecSrc + gpGlobals->v_forward * (pev->scale * 8.0f);
	UTIL_TraceLine( vecSrc, vecEnd, ignore_monsters, edict(), &tr );

	// this is our second PVS point
	CBaseEntity *pVisHelper = CBaseEntity::Instance( pev->enemy );
	if( pVisHelper ) UTIL_SetOrigin( pVisHelper, tr.vecEndPos + tr.vecPlaneNormal * 8.0f ); 
}

// =================== ENV_DYNLIGHT ==============================================

#define SF_DYNLIGHT_START_ON		BIT( 0 )
#define SF_DYNLIGHT_NOLIGHT_IN_SOLID	BIT( 2 )

class CEnvDynamicLight : public CBaseDelay
{
	DECLARE_CLASS( CEnvDynamicLight, CBaseDelay );
public: 
	void Spawn( void );
	void KeyValue( KeyValueData *pkvd );
	void Use( CBaseEntity *pActivator, CBaseEntity *pCaller, USE_TYPE useType, float value );
	virtual STATE GetState( void ) { return FBitSet( pev->effects, EF_NODRAW ) ? STATE_OFF : STATE_ON; };
	virtual int ObjectCaps( void ) { return BaseClass :: ObjectCaps() & ~FCAP_ACROSS_TRANSITION; }
};

LINK_ENTITY_TO_CLASS( light_dynamic, CEnvDynamicLight );

void CEnvDynamicLight :: KeyValue( KeyValueData *pkvd )
{
	if( FStrEq( pkvd->szKeyName, "radius" ))
	{
		pev->scale = Q_atof( pkvd->szValue ) * (1.0f/8.0f);
		pkvd->fHandled = TRUE;
	}
	else if( FStrEq( pkvd->szKeyName, "style" ))
	{
		pev->renderfx = Q_atoi( pkvd->szValue );
		pkvd->fHandled = TRUE;
	}
	else BaseClass::KeyValue( pkvd );
}

void CEnvDynamicLight::Spawn( void )
{
	pev->effects |= (EF_DYNAMIC_LIGHT|EF_NODRAW);
	pev->rendermode = kRenderTransTexture;
	pev->movetype = MOVETYPE_NOCLIP;
	pev->renderamt = 0;

	if( !pev->scale )
		pev->scale = 300 * (1.0f/8.0f);

	pev->frags = pev->scale; // member radius

	SET_MODEL( edict(), "sprites/null.spr" );
	SetBits( m_iFlags, MF_POINTENTITY );
//	SetLocalAvelocity( Vector( 30, 30, 30 ));
	RelinkEntity( FALSE );

	if( FBitSet( pev->spawnflags, SF_PROJECTOR_NO_ATTEN ))
		pev->iuser1 |= CF_NOATTEN;

	if( FBitSet( pev->spawnflags, SF_DYNLIGHT_NOLIGHT_IN_SOLID ))
		pev->iuser1 |= CF_NOLIGHT_IN_SOLID;

	// to prevent disapperaing from PVS (scale is premultiplied by 0.125)
	UTIL_SetSize( pev, Vector( -pev->scale, -pev->scale, -pev->scale ), Vector( pev->scale, pev->scale, pev->scale ));

	// enable dynlight
	if( FStringNull( pev->targetname ) || FBitSet( pev->spawnflags, SF_DYNLIGHT_START_ON ))
	{
		SetThink( SUB_CallUseToggle );
		SetNextThink( 0.1 );
	}
}

void CEnvDynamicLight::Use( CBaseEntity *pActivator, CBaseEntity *pCaller, USE_TYPE useType, float value )
{
	if( IsLockedByMaster( pActivator ))
		return;

 	if( useType == USE_SET )
 	{
 		// set radius
 		value = bound( 0.0f, value, 1.0f );
		pev->scale = pev->frags * value;
		return;
	}

	if( FBitSet( pev->effects, EF_NODRAW ))
		pev->effects &= ~EF_NODRAW;
	else pev->effects |= EF_NODRAW;
}

// =================== ENV_PARTICLE ==============================================

#define SF_PARTICLE_START_ON		BIT( 0 )

class CBaseParticle : public CPointEntity
{
	DECLARE_CLASS( CBaseParticle, CPointEntity );
public:
	void Spawn( void );
	void KeyValue( KeyValueData *pkvd );
	void Use( CBaseEntity *pActivator, CBaseEntity *pCaller, USE_TYPE useType, float value );
	virtual STATE GetState( void ) { return (pev->renderfx == kRenderFxAurora) ? STATE_ON : STATE_OFF; };
	void StartMessage( CBasePlayer *pPlayer );
};

LINK_ENTITY_TO_CLASS( env_particle, CBaseParticle );

void CBaseParticle :: KeyValue( KeyValueData *pkvd )
{
	if( FStrEq( pkvd->szKeyName, "aurora" ))
	{
		pev->message = ALLOC_STRING( pkvd->szValue );
		pkvd->fHandled = TRUE;
	}
	else if( FStrEq( pkvd->szKeyName, "attachment" ))
	{
		pev->impulse = Q_atoi( pkvd->szValue );
		pev->impulse = bound( 1, pev->impulse, 4 );
		pkvd->fHandled = TRUE;
	}
	else BaseClass::KeyValue( pkvd );
}

void CBaseParticle :: StartMessage( CBasePlayer *pPlayer )
{
	MESSAGE_BEGIN( MSG_ONE, gmsgParticle, NULL, pPlayer->pev );
		WRITE_ENTITY( entindex() );
		WRITE_STRING( STRING( pev->message ));
		WRITE_BYTE( pev->impulse );
	MESSAGE_END();
}

void CBaseParticle::Spawn( void )
{
	SET_MODEL( edict(), "sprites/null.spr" );
	UTIL_SetSize( pev, Vector( -8, -8, -8 ), Vector( 8, 8, 8 ));
	RelinkEntity();

	if( FBitSet( pev->spawnflags, SF_PARTICLE_START_ON ))
		pev->renderfx = kRenderFxAurora;
}

void CBaseParticle::Use( CBaseEntity *pActivator, CBaseEntity *pCaller, USE_TYPE useType, float value )
{
	if( useType == USE_TOGGLE )
	{
		if( GetState() == STATE_ON )
			useType = USE_OFF;
		else useType = USE_ON;
	}

	if( useType == USE_ON )
	{
		pev->renderfx = kRenderFxAurora;
	}
	else if( useType == USE_OFF )
	{
		pev->renderfx = kRenderFxNone;
	}
}

// =================== ENV_COUNTER ==============================================

class CDecLED : public CBaseDelay
{
	DECLARE_CLASS( CDecLED, CBaseDelay );
public:
	void	Spawn( void );
	void	Precache( void );
	void	CheckState( void );
	void	KeyValue( KeyValueData *pkvd );
	void	Use( CBaseEntity *pActivator, CBaseEntity *pCaller, USE_TYPE useType, float value );
	virtual int ObjectCaps( void ) { return BaseClass :: ObjectCaps() & ~FCAP_ACROSS_TRANSITION; }
	float 	Frames( void ) { return MODEL_FRAMES( pev->modelindex ) - 1; }
	float	curframe( void ) { return Q_rint( pev->frame ); }
private:
	bool	keyframeset;
	float	flashtime;
	bool	recursive; // to prevent infinite loop with USE_RESET
};

LINK_ENTITY_TO_CLASS( env_counter, CDecLED );

void CDecLED :: Precache( void )
{
	if (pev->model)
		PRECACHE_MODEL((char*)STRING(pev->model)); //LRC
	else
		PRECACHE_MODEL("sprites/decimal.spr");
}

void CDecLED :: Spawn( void )
{
	Precache( );

	if (pev->model)
		SET_MODEL(ENT(pev), STRING(pev->model)); //LRC
	else
		SET_MODEL(ENT(pev), "sprites/decimal.spr");
         
	// Smart Field System 
          if( !pev->renderamt ) pev->renderamt = 200; // light transparency
	if( !pev->rendermode ) pev->rendermode = kRenderTransAdd;
	if( !pev->frags ) pev->frags = Frames();
	if( !keyframeset ) pev->impulse = -1;

	CheckState();

	SetBits( m_iFlags, MF_POINTENTITY );
	pev->effects |= EF_NOINTERP;
	pev->movetype = MOVETYPE_NONE;
	pev->solid = SOLID_NOT;

	Vector angles = GetLocalAngles();

	angles.x = -angles.x;
	angles.y = 180 - angles.y;
	angles.y = -angles.y;

	SetLocalAngles( angles );	
	RelinkEntity( FALSE );
	pev->framerate = -1.0f;
	pev->button = 1;
}

void CDecLED :: CheckState( void )
{
	switch( (int)pev->dmg )
	{
	case 1:
		if( pev->impulse >= curframe() )
			m_iState = STATE_ON;
		else m_iState = STATE_OFF;
		break;
	case 2:
		if( pev->impulse <= curframe() )
			m_iState = STATE_ON;
		else m_iState = STATE_OFF;
		break;
	default:
		if( pev->impulse == curframe() )
			m_iState = STATE_ON;
		else m_iState = STATE_OFF;
		break;
	}
}

void CDecLED :: KeyValue( KeyValueData *pkvd )
{
	if( FStrEq( pkvd->szKeyName, "maxframe" ))
	{
		pev->frags = Q_atof( pkvd->szValue );
		pkvd->fHandled = TRUE;
	}
	else if( FStrEq( pkvd->szKeyName, "keyframe" ))
	{
		pev->impulse = Q_atoi( pkvd->szValue );
		pkvd->fHandled = TRUE;
		keyframeset = TRUE;
	}
	else if( FStrEq( pkvd->szKeyName, "type" ))
	{
		pev->dmg = Q_atoi( pkvd->szValue );
		pkvd->fHandled = TRUE;
	}
	else BaseClass::KeyValue( pkvd );
}

void CDecLED :: Use( CBaseEntity *pActivator, CBaseEntity *pCaller, USE_TYPE useType, float value )
{
	if( useType == USE_ON || useType == USE_TOGGLE )
	{
		if( pev->frame >= pev->frags ) 
		{
			UTIL_FireTargets( pev->target, pActivator, this, useType, value );
			pev->frame = 0;
		}
		else pev->frame++;
		
		pev->button = 1;
		ClearBits( pev->effects, EF_NODRAW );
		flashtime = gpGlobals->time + 0.5f;
	}
	else if( useType == USE_OFF )
	{
		if( pev->frame <= 0 ) 
		{
         			UTIL_FireTargets( pev->target, pActivator, this, useType, value );
			pev->frame = pev->frags;
		}
		else pev->frame--;

		pev->button = 0;
		ClearBits( pev->effects, EF_NODRAW );
		flashtime = gpGlobals->time + 0.5f;
	}
	else if( useType == USE_SET ) // set custom frame
	{ 
		if( value != 0.0f )
		{
			if( value == -1 ) value = 0; // reset frame
			pev->frame = fmod( value, pev->frags + 1 );
                             	float next = value / ( pev->frags + 1 );
                             	if( next ) UTIL_FireTargets( pev->target, pActivator, this, useType, next );
	         		ClearBits( pev->effects, EF_NODRAW );
			flashtime = gpGlobals->time + 0.5f;
			pev->button = 1;
		}
		else if( gpGlobals->time > flashtime )
		{
			if( pev->button )
			{
				SetBits( pev->effects, EF_NODRAW );
				pev->button = 0;
			}
			else
			{
				ClearBits( pev->effects, EF_NODRAW );
				pev->button = 1;
			}
		}
	}
	else if( useType == USE_RESET ) // immediately reset
	{ 
		pev->frame = 0;
		pev->button = 1;
		pev->effects &= ~EF_NODRAW;

		if( !recursive )
		{
			recursive = TRUE;
			UTIL_FireTargets( pev->target, pActivator, this, useType, 0 );
			recursive = FALSE;
		}
		else ALERT( at_error, "%s has infinite loop in chain\n", STRING( pev->classname ));
	}

	// check state after changing
	CheckState();
}


#define SF_REMOVE_ON_FIRE		0x0001
#define SF_KILL_CENTER		0x0002

class CEnvWarpBall : public CBaseEntity
{
	DECLARE_CLASS( CEnvWarpBall, CBaseEntity );
public:
	void Precache( void );
	void Spawn( void ) { Precache(); }
	void Think( void );
	void KeyValue( KeyValueData *pkvd );
	void Use( CBaseEntity *pActivator, CBaseEntity *pCaller, USE_TYPE useType, float value );
	virtual int ObjectCaps( void ) { return BaseClass :: ObjectCaps() & ~FCAP_ACROSS_TRANSITION; }
	Vector vecOrigin;
};

LINK_ENTITY_TO_CLASS( env_warpball, CEnvWarpBall );

void CEnvWarpBall :: KeyValue( KeyValueData *pkvd )
{
	if (FStrEq(pkvd->szKeyName, "radius"))
	{
		pev->button = atoi(pkvd->szValue);
		pkvd->fHandled = TRUE;
	}
	if (FStrEq(pkvd->szKeyName, "warp_target"))
	{
		pev->message = ALLOC_STRING(pkvd->szValue);
		pkvd->fHandled = TRUE;
	}
	if (FStrEq(pkvd->szKeyName, "damage_delay"))
	{
		pev->frags = atof(pkvd->szValue);
		pkvd->fHandled = TRUE;
	}
	else
		CBaseEntity::KeyValue( pkvd );
}

void CEnvWarpBall::Precache( void )
{
	PRECACHE_MODEL( "sprites/beam/lgtning.spr" );
	PRECACHE_MODEL( "sprites/explode/Fexplo1.spr" );
	PRECACHE_SOUND( "debris/beam/beamstart2.wav" );
	PRECACHE_SOUND( "debris/beam/beamstart7.wav" );
}

void CEnvWarpBall::Use( CBaseEntity *pActivator, CBaseEntity *pCaller, USE_TYPE useType, float value )
{
	int iTimes = 0;
	int iDrawn = 0;
	TraceResult tr;
	Vector vecDest;
	CBeam *pBeam;
	CBaseEntity *pEntity = UTIL_FindEntityByTargetname ( NULL, STRING(pev->message));
	edict_t *pos;
		
	if( pEntity ) // target found ?
	{
		vecOrigin = pEntity->GetAbsOrigin();
		pos = pEntity->edict();
	}
	else
	{         //use as center
		vecOrigin = GetAbsOrigin();
		pos = edict();
	}	

	EMIT_SOUND( pos, CHAN_BODY, "debris/beamstart2.wav", 1, ATTN_NORM );
	UTIL_ScreenShake( vecOrigin, 6, 160, 1.0, pev->button );
	CSprite *pSpr = CSprite::SpriteCreate( "sprites/explode/Fexplo1.spr", vecOrigin, TRUE );
	pSpr->SetParent( this );
	pSpr->AnimateAndDie( 18 );
	pSpr->SetTransparency(kRenderGlow,  77, 210, 130,  255, kRenderFxNoDissipation);
	EMIT_SOUND( pos, CHAN_ITEM, "debris/beamstart7.wav", 1, ATTN_NORM );
	int iBeams = RANDOM_LONG(20, 40);
	while (iDrawn < iBeams && iTimes < (iBeams * 3))
	{
		vecDest = pev->button * (Vector(RANDOM_FLOAT(-1,1), RANDOM_FLOAT(-1,1), RANDOM_FLOAT(-1,1)).Normalize());
		UTIL_TraceLine( vecOrigin, vecOrigin + vecDest, ignore_monsters, NULL, &tr);
		if (tr.flFraction != 1.0)
		{
			// we hit something.
			iDrawn++;
			pBeam = CBeam::BeamCreate("sprites/beam/lgtning.spr", 200);
			pBeam->PointsInit( vecOrigin, tr.vecEndPos );
			pBeam->SetColor( 20, 243, 20 );
			pBeam->SetNoise( 65 );
			pBeam->SetBrightness( 220 );
			pBeam->SetWidth( 30 );
			pBeam->SetScrollRate( 35 );
//			pBeam->SetParent( this );
			pBeam->SetThink(&CBeam:: SUB_Remove );
			pBeam->pev->nextthink = gpGlobals->time + RANDOM_FLOAT(0.5, 1.6);
		}
		iTimes++;
	}
	pev->nextthink = gpGlobals->time + pev->frags;
}

void CEnvWarpBall::Think( void )
{
	SUB_UseTargets( this, USE_TOGGLE, 0);
 
	if ( pev->spawnflags & SF_KILL_CENTER )
	{
		CBaseEntity *pMonster = NULL;

		while ((pMonster = UTIL_FindEntityInSphere( pMonster, vecOrigin, 72 )) != NULL)
		{
			if ( FBitSet( pMonster->pev->flags, FL_MONSTER ) || FClassnameIs( pMonster->pev, "player"))
				pMonster->TakeDamage ( pev, pev, 100, DMG_GENERIC );
		}
	}
	if ( pev->spawnflags & SF_REMOVE_ON_FIRE ) UTIL_Remove( this );
}
